# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import io
import os
import optparse
import re
import shutil
import sys
import textwrap
import logging
from logging.config import fileConfig

from mapproxy.config.loader import load_plugins
from mapproxy.script.conf.app import config_command
from mapproxy.script.defrag import defrag_command
from mapproxy.script.export import export_command
from mapproxy.script.grids import grids_command
from mapproxy.script.scales import scales_command
from mapproxy.script.wms_capabilities import wms_capabilities_command
from mapproxy.script.gridconf_from_ogcapitilematrixset import gridconf_from_ogcapitilematrixset_command
from mapproxy.version import version


def setup_logging(level=logging.INFO, format=None):
    mapproxy_log = logging.getLogger('mapproxy')
    mapproxy_log.setLevel(level)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(level)
    if not format:
        format = "[%(asctime)s] %(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(format)
    ch.setFormatter(formatter)
    mapproxy_log.addHandler(ch)


def serve_develop_command(args):
    parser = optparse.OptionParser("usage: %prog serve-develop [options] mapproxy.yaml")
    parser.add_option("-b", "--bind",
                      dest="address", default='127.0.0.1:8080',
                      help="Server socket [127.0.0.1:8080]. Use 0.0.0.0 for external access. :1234 to change port.")
    parser.add_option("--debug", default=False, action='store_true',
                      dest="debug",
                      help="Enable debug mode")
    parser.add_option("--log-config", dest="log_config", help="Path to a log config file")
    options, args = parser.parse_args(args)

    if len(args) != 2:
        parser.print_help()
        print("\nERROR: MapProxy configuration required.")
        sys.exit(1)

    mapproxy_conf = args[1]

    host, port = parse_bind_address(options.address)

    if options.debug and host not in ('localhost', '127.0.0.1'):
        print(textwrap.dedent("""\
        ################# WARNING! ##################
        Running debug mode with non-localhost address
        is a serious security vulnerability.
        #############################################\
        """))

    if options.debug:
        setup_logging(level=logging.DEBUG)
    elif options.log_config is not None:
        fileConfig(options.log_config)
    else:
        setup_logging()
    from mapproxy.wsgiapp import make_wsgi_app
    from mapproxy.config.loader import ConfigurationError
    from mapproxy.util.ext.serving import run_simple
    try:
        app = make_wsgi_app(mapproxy_conf, debug=options.debug)
    except ConfigurationError:
        sys.exit(2)

    extra_files = app.config_files.keys()

    if options.debug:
        try:
            from repoze.profile import ProfileMiddleware
            app = ProfileMiddleware(
                app,
                log_filename='/tmp/mapproxy_profile.log',
                discard_first_request=True,
                flush_at_shutdown=True,
                path='/__profile__',
                unwind=False,
            )
            print('Installed profiler at /__profile__')
        except ImportError:
            pass

    run_simple(host, port, app, use_reloader=True, processes=1,
               threaded=True, passthrough_errors=True,
               extra_files=extra_files)


def serve_multiapp_develop_command(args):
    parser = optparse.OptionParser("usage: %prog serve-multiapp-develop [options] projects/")
    parser.add_option("-b", "--bind",
                      dest="address", default='127.0.0.1:8080',
                      help="Server socket [127.0.0.1:8080]")
    parser.add_option("--debug", default=False, action='store_true',
                      dest="debug",
                      help="Enable debug mode")
    options, args = parser.parse_args(args)

    if len(args) != 2:
        parser.print_help()
        print("\nERROR: MapProxy projects directory required.")
        sys.exit(1)

    mapproxy_conf_dir = args[1]

    host, port = parse_bind_address(options.address)

    if options.debug and host not in ('localhost', '127.0.0.1'):
        print(textwrap.dedent("""\
        ################# WARNING! ##################
        Running debug mode with non-localhost address
        is a serious security vulnerability.
        #############################################\
        """))

    setup_logging()
    from mapproxy.multiapp import make_wsgi_app
    from mapproxy.util.ext.serving import run_simple
    app = make_wsgi_app(mapproxy_conf_dir, debug=options.debug)

    run_simple(host, port, app, use_reloader=True, processes=1,
               threaded=True, passthrough_errors=True)


def parse_bind_address(address, default=('localhost', 8080)):
    """
    >>> parse_bind_address('80')
    ('localhost', 80)
    >>> parse_bind_address('0.0.0.0')
    ('0.0.0.0', 8080)
    >>> parse_bind_address('0.0.0.0:8081')
    ('0.0.0.0', 8081)
    """
    if ':' in address:
        host, port = address.split(':', 1)
        port = int(port)
    elif re.match(r'^\d+$', address):
        host = default[0]
        port = int(address)
    else:
        host = address
        port = default[1]
    return host, port


def create_command(args):
    cmd = CreateCommand(args)
    cmd.run()


class CreateCommand(object):
    templates = {
        'base-config': {},
        'wsgi-app': {},
        'log-ini': {},
    }

    def __init__(self, args):
        parser = optparse.OptionParser("usage: %prog create [options] [destination]")
        parser.add_option("-t", "--template", dest="template",
                          help="Create a configuration from this template.")
        parser.add_option("-l", "--list-templates", dest="list_templates",
                          action="store_true", default=False,
                          help="List all available configuration templates.")
        parser.add_option("-f", "--mapproxy-conf", dest="mapproxy_conf",
                          help="Existing MapProxy configuration (required for some templates).")
        parser.add_option("--force", dest="force", action="store_true",
                          default=False, help="Force operation (e.g. overwrite existing files).")

        self.options, self.args = parser.parse_args(args)
        self.parser = parser

    def log_error(self, msg, *args):
        print('ERROR:', msg % args, file=sys.stderr)

    def run(self):

        if self.options.list_templates:
            print_items(self.templates, title="Available templates")
            sys.exit(1)
        elif self.options.template:
            if self.options.template not in self.templates:
                self.log_error("unknown template " + self.options.template)
                sys.exit(1)

            if len(self.args) != 2:
                self.log_error("template requires destination argument")
                sys.exit(1)

            sys.exit(
                getattr(self, 'template_' + self.options.template.replace('-', '_'))()
            )
        else:
            self.parser.print_help()
            sys.exit(1)

    @property
    def mapproxy_conf(self):
        if not self.options.mapproxy_conf:
            self.parser.print_help()
            self.log_error("template requires --mapproxy-conf option")
            sys.exit(1)
        return os.path.abspath(self.options.mapproxy_conf)

    def template_dir(self):
        import mapproxy.config_template
        template_dir = os.path.join(
            os.path.dirname(mapproxy.config_template.__file__),
            'base_config')
        return template_dir

    def template_wsgi_app(self):
        app_filename = self.args[1]
        if '.' not in os.path.basename(app_filename):
            app_filename += '.py'
        mapproxy_conf = self.mapproxy_conf
        if os.path.exists(app_filename) and not self.options.force:
            self.log_error("%s already exists, use --force", app_filename)
            return 1

        print("writing MapProxy app to %s" % (app_filename, ))

        template_dir = self.template_dir()
        app_template = io.open(os.path.join(template_dir, 'config.wsgi'), encoding='utf-8').read()
        with io.open(app_filename, 'w', encoding='utf-8') as f:
            f.write(app_template % {'mapproxy_conf': mapproxy_conf,
                                    'here': os.path.dirname(mapproxy_conf)})

        return 0

    def template_base_config(self):
        outdir = self.args[1]
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        template_dir = self.template_dir()

        for filename in ('mapproxy.yaml', 'seed.yaml',
                         'full_example.yaml', 'full_seed_example.yaml'):
            to = os.path.join(outdir, filename)
            from_ = os.path.join(template_dir, filename)
            if os.path.exists(to) and not self.options.force:
                self.log_error("%s already exists, use --force", to)
                return 1
            print("writing %s" % (to, ))
            shutil.copy(from_, to)

        return 0

    def template_log_ini(self):
        log_filename = self.args[1]

        if os.path.exists(log_filename) and not self.options.force:
            self.log_error("%s already exists, use --force", log_filename)
            return 1

        template_dir = self.template_dir()
        log_template = io.open(os.path.join(template_dir, 'log.ini'), encoding='utf-8').read()
        with io.open(log_filename, 'w', encoding='utf-8') as f:
            f.write(log_template)

        return 0


commands = {
    'serve-develop': {
        'func': serve_develop_command,
        'help': 'Run MapProxy development server.'
    },
    'serve-multiapp-develop': {
        'func': serve_multiapp_develop_command,
        'help': 'Run MultiMapProxy development server.'
    },
    'create': {
        'func': create_command,
        'help': 'Create example configurations.'
    },
    'scales': {
        'func': scales_command,
        'help': 'Convert between scales and resolutions.'
    },
    'wms-capabilities': {
        'func': wms_capabilities_command,
        'help': 'Display WMS capabilites.',
    },
    'grids': {
        'func': grids_command,
        'help': 'Display detailed informations for configured grids.'
    },
    'export': {
        'func': export_command,
        'help': 'Export existing caches.'
    },
    'autoconfig': {
        'func': config_command,
        'help': 'Create config from WMS capabilities or a Geopackage file.'
    },
    'defrag-compact-cache': {
        'func': defrag_command,
        'help': 'De-fragmentate compact caches.'
    },
    'gridconf-from-ogcapitilematrixset': {
        'func': gridconf_from_ogcapitilematrixset_command,
        'help': 'Export OGC API TileMatrixSet as MapProxy grid configuration.'
    },
}


def register_command(command_name, command_spec):
    """ Method used by plugins to register a command.

        :param command_name: Name of the service
        :type command_name: str
        :param command_spec: Definition of the command. Dictionary with a 'func' and 'help' member
        :type command_spec: dict
    """

    commands[command_name] = command_spec


class NonStrictOptionParser(optparse.OptionParser):
    def _process_args(self, largs, rargs, values):
        while rargs:
            arg = rargs[0]
            # We handle bare "--" explicitly, and bare "-" is handled by the
            # standard arg handler since the short arg case ensures that the
            # len of the opt string is greater than 1.
            try:
                if arg == "--":
                    del rargs[0]
                    return
                elif arg[0:2] == "--":
                    # process a single long option (possibly with value(s))
                    self._process_long_opt(rargs, values)
                elif arg[:1] == "-" and len(arg) > 1:
                    # process a cluster of short options (possibly with
                    # value(s) for the last one only)
                    self._process_short_opts(rargs, values)
                elif self.allow_interspersed_args:
                    largs.append(arg)
                    del rargs[0]
                else:
                    return
            except optparse.BadOptionError:
                largs.append(arg)


def print_items(data, title='Commands'):
    name_len = max(len(name) for name in data)

    if title:
        print('%s:' % (title, ), file=sys.stdout)
    for name, item in data.items():
        help = item.get('help', '')
        name = ('%%-%ds' % name_len) % name
        if help:
            help = '  ' + help
        print('  %s%s' % (name, help), file=sys.stdout)


def main():

    load_plugins()

    parser = NonStrictOptionParser("usage: %prog COMMAND [options]",
                                   add_help_option=False)
    options, args = parser.parse_args()

    if len(args) < 1 or args[0] in ('--help', '-h'):
        parser.print_help()
        print()
        print_items(commands)
        sys.exit(1)

    if len(args) == 1 and args[0] == '--version':
        print('MapProxy ' + version)
        sys.exit(1)

    command = args[0]
    if command not in commands:
        parser.print_help()
        print()
        print_items(commands)
        print('\nERROR: unknown command %s' % (command,), file=sys.stdout)
        sys.exit(1)

    args = sys.argv[0:1] + sys.argv[2:]
    commands[command]['func'](args)


if __name__ == '__main__':
    main()



# -:- encoding: utf8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from PIL import Image, ImageDraw, ImageColor, ImageFont
from mapproxy.cache.tile import Tile
from mapproxy.image import ImageSource
from mapproxy.image.message import TextDraw, message_image
from mapproxy.image.opts import ImageOptions
from mapproxy.tilefilter import watermark_filter

import pytest

PNG_FORMAT = ImageOptions(format="image/png")

requires_freetype = pytest.mark.skipif(
    not isinstance(ImageFont.load_default(), ImageFont.FreeTypeFont),
    reason="Test expects the default Pillow FreeTypeFont")


class TestTextDraw(object):

    def test_ul(self):
        font = ImageFont.load_default()
        td = TextDraw("Hello", font)
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        total_box, boxes = td.text_boxes(draw, (100, 100))
        assert total_box == boxes[0]
        assert len(boxes) == 1

    @requires_freetype
    def test_multiline_ul(self):
        font = ImageFont.load_default()
        td = TextDraw("Hello\nWorld", font)
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        total_box, boxes = td.text_boxes(draw, (100, 100))
        assert total_box == (5, 7, 33, 28)
        assert boxes == [(5, 7, 30, 15), (5, 20, 33, 28)]

    @requires_freetype
    def test_multiline_lr(self):
        font = ImageFont.load_default()
        td = TextDraw("Hello\nWorld", font, placement="lr")
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        total_box, boxes = td.text_boxes(draw, (100, 100))
        assert total_box == (67, 76, 95, 97)
        assert boxes == [(67, 76, 92, 84), (67, 89, 95, 97)]

    @requires_freetype
    def test_multiline_center(self):
        font = ImageFont.load_default()
        td = TextDraw("Hello\nWorld", font, placement="cc")
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        total_box, boxes = td.text_boxes(draw, (100, 100))
        assert total_box == (36, 42, 64, 63)
        assert boxes == [(36, 42, 61, 50), (36, 55, 64, 63)]

    @requires_freetype
    def test_unicode(self):
        font = ImageFont.load_default()
        td = TextDraw(u"Héllö\nWørld", font, placement="cc")
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        total_box, boxes = td.text_boxes(draw, (100, 100))
        assert total_box == (36, 42, 64, 63)
        assert boxes == [(36, 42, 60, 50), (36, 55, 64, 63)]

    def _test_all(self):
        for x in "c":
            for y in "LR":
                yield self.check_placement, x, y

    def check_placement(self, x, y):
        font = ImageFont.load_default()
        td = TextDraw(
            "Hello\nWorld\n%s %s" % (x, y),
            font,
            placement=x + y,
            padding=5,
            linespacing=2,
        )
        img = Image.new("RGB", (100, 100))
        draw = ImageDraw.Draw(img)
        td.draw(draw, img.size)
        img.show()

    def test_transparent(self):
        font = ImageFont.load_default()
        td = TextDraw("Hello\nWorld", font, placement="cc")
        img = Image.new("RGBA", (100, 100), (0, 0, 0, 0))
        draw = ImageDraw.Draw(img)
        td.draw(draw, img.size)
        # override the alpha value as pillow >= 10.1.0 uses a new default font with transparency
        img.putalpha(255)

        assert len(img.getcolors()) == 2
        # top color (bg) is transparent
        assert sorted(img.getcolors())[1][1] == (0, 0, 0, 255)


class TestMessageImage(object):

    def test_blank(self):
        image_opts = PNG_FORMAT.copy()
        image_opts.bgcolor = "#113399"
        img = message_image("", size=(100, 150), image_opts=image_opts)
        assert isinstance(img, ImageSource)
        assert img.size == (100, 150)
        pil_img = img.as_image()
        assert pil_img.getpixel((0, 0)) == ImageColor.getrgb("#113399")
        # 3 values in histogram (RGB)
        assert [x for x in pil_img.histogram() if x > 0] == [
            15000,
            15000,
            15000,
        ]

    @requires_freetype
    def test_message(self):
        image_opts = PNG_FORMAT.copy()
        image_opts.bgcolor = "#113399"
        img = message_image("test", size=(100, 150), image_opts=image_opts)
        text_pixels = 75
        image_pixels = 100 * 150
        assert isinstance(img, ImageSource)
        assert img.size == (100, 150)
        # expect the large histogram count values to be the amount of background pixels
        assert [x for x in img.as_image().histogram() if x > 10] == [
            image_pixels - text_pixels,
            image_pixels - text_pixels,
            image_pixels - text_pixels,
        ]

    def test_transparent(self):
        image_opts = ImageOptions(transparent=True)
        print(image_opts)
        img = message_image("", size=(100, 150), image_opts=image_opts)
        assert isinstance(img, ImageSource)
        assert img.size == (100, 150)
        pil_img = img.as_image()
        assert pil_img.getpixel((0, 0)) == (255, 255, 255, 0)
        # 6 values in histogram (3xRGB for background, 3xRGB for text message)
        assert [x for x in pil_img.histogram() if x > 0] == [
            15000,
            15000,
            15000,
            15000,
        ]


class TestWatermarkTileFilter(object):

    def setup_method(self):
        self.tile = Tile((0, 0, 0))
        self.filter = watermark_filter("Test")

    def test_filter(self):
        img = Image.new("RGB", (200, 200))
        orig_source = ImageSource(img)
        self.tile.source = orig_source
        filtered_tile = self.filter(self.tile)

        assert self.tile is filtered_tile
        assert orig_source != filtered_tile.source

        pil_img = filtered_tile.source.as_image()
        assert pil_img.getpixel((0, 0)) == (0, 0, 0)

        colors = pil_img.getcolors()
        colors.sort()
        # most but not all parts are bg color
        assert 39950 > colors[-1][0] > 39000
        assert colors[-1][1] == (0, 0, 0)

    def test_filter_with_alpha(self):
        img = Image.new("RGBA", (200, 200), (10, 15, 20, 0))
        orig_source = ImageSource(img)
        self.tile.source = orig_source
        filtered_tile = self.filter(self.tile)

        assert self.tile is filtered_tile
        assert orig_source != filtered_tile.source

        pil_img = filtered_tile.source.as_image()
        assert pil_img.getpixel((0, 0)) == (10, 15, 20, 0)

        colors = pil_img.getcolors()
        colors.sort()
        # most but not all parts are bg color
        assert 39950 > colors[-1][0] > 39000
        assert colors[-1][1] == (10, 15, 20, 0)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from mapproxy.image import ImageSource
from mapproxy.image.transform import ImageTransformer
from mapproxy.image.opts import create_image

import logging
log = logging.getLogger(__name__)


class TileMerger(object):
    """
    Merge multiple tiles into one image.
    """

    def __init__(self, tile_grid, tile_size):
        """
        :param tile_grid: the grid size
        :type tile_grid: ``(int(x_tiles), int(y_tiles))``
        :param tile_size: the size of each tile
        """
        self.tile_grid = tile_grid
        self.tile_size = tile_size

    def merge(self, ordered_tiles, image_opts):
        """
        Merge all tiles into one image.

        :param ordered_tiles: list of tiles, sorted row-wise (top to bottom)
        :rtype: `ImageSource`
        """
        if self.tile_grid == (1, 1):
            assert len(ordered_tiles) == 1
            if ordered_tiles[0] is not None:
                tile = ordered_tiles.pop()
                return tile
        src_size = self._src_size()

        result = create_image(src_size, image_opts)

        cacheable = True

        for i, source in enumerate(ordered_tiles):
            if source is None:
                continue
            try:
                if not source.cacheable:
                    cacheable = False
                tile = source.as_image()
                pos = self._tile_offset(i)
                tile.draft(image_opts.mode, self.tile_size)
                result.paste(tile, pos)
                source.close_buffers()
            except IOError as e:
                if e.errno is None:  # PIL error
                    log.warning('unable to load tile %s, removing it (reason was: %s)'
                                % (source, str(e)))
                    if getattr(source, 'filename'):
                        if os.path.exists(source.filename):
                            os.remove(source.filename)
                else:
                    raise
        return ImageSource(result, size=src_size, image_opts=image_opts, cacheable=cacheable)

    def _src_size(self):
        width = self.tile_grid[0]*self.tile_size[0]
        height = self.tile_grid[1]*self.tile_size[1]
        return width, height

    def _tile_offset(self, i):
        """
        Return the image offset (upper-left coord) of the i-th tile,
        where the tiles are ordered row-wise, top to bottom.
        """
        return (i % self.tile_grid[0]*self.tile_size[0],
                i//self.tile_grid[0]*self.tile_size[1])


class TileSplitter(object):
    """
    Splits a large image into multiple tiles.
    """

    def __init__(self, meta_tile, image_opts):
        self.meta_img = meta_tile.as_image()
        self.image_opts = image_opts

    def get_tile(self, crop_coord, tile_size):
        """
        Return the cropped tile.
        :param crop_coord: the upper left pixel coord to start
        :param tile_size: width and height of the new tile
        :rtype: `ImageSource`
        """
        minx, miny = crop_coord
        maxx = minx + tile_size[0]
        maxy = miny + tile_size[1]

        if (minx < 0 or miny < 0 or maxx > self.meta_img.size[0]
                or maxy > self.meta_img.size[1]):

            crop = self.meta_img.crop((
                max(minx, 0),
                max(miny, 0),
                min(maxx, self.meta_img.size[0]),
                min(maxy, self.meta_img.size[1])))
            result = create_image(tile_size, self.image_opts)
            result.paste(crop, (abs(min(minx, 0)), abs(min(miny, 0))))
            crop = result
        else:
            crop = self.meta_img.crop((minx, miny, maxx, maxy))
        return ImageSource(crop, size=tile_size, image_opts=self.image_opts)


class TiledImage(object):
    """
    An image built-up from multiple tiles.
    """

    def __init__(self, tiles, tile_grid, tile_size, src_bbox, src_srs):
        """
        :param tiles: all tiles (sorted row-wise, top to bottom)
        :param tile_grid: the tile grid size
        :type tile_grid: ``(int(x_tiles), int(y_tiles))``
        :param tile_size: the size of each tile
        :param src_bbox: the bbox of all tiles
        :param src_srs: the srs of the bbox
        :param transparent: if the sources are transparent
        """
        self.tiles = tiles
        self.tile_grid = tile_grid
        self.tile_size = tile_size
        self.src_bbox = src_bbox
        self.src_srs = src_srs

    def image(self, image_opts):
        """
        Return the tiles as one merged image.

        :rtype: `ImageSource`
        """
        tm = TileMerger(self.tile_grid, self.tile_size)
        return tm.merge(self.tiles, image_opts=image_opts)

    def transform(self, req_bbox, req_srs, out_size, image_opts):
        """
        Return the the tiles as one merged and transformed image.

        :param req_bbox: the bbox of the output image
        :param req_srs: the srs of the req_bbox
        :param out_size: the size in pixel of the output image
        :rtype: `ImageSource`
        """
        transformer = ImageTransformer(self.src_srs, req_srs)
        src_img = self.image(image_opts)
        return transformer.transform(src_img, self.src_bbox, out_size, req_bbox,
                                     image_opts)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import datetime

from mapproxy.util.ext.dictspec.validator import validate, ValidationError
from mapproxy.util.ext.dictspec.spec import one_of, anything, number
from mapproxy.util.ext.dictspec.spec import recursive, required, type_spec, combined


def validate_options(conf_dict):
    """
    Validate `conf_dict` agains mapproxy.yaml spec.
    Returns tuple with a list of errors and a bool.
    The list is empty when no errors where found.
    The bool is True when the errors are informal and not critical.
    """
    try:
        validate(mapproxy_yaml_spec, conf_dict)
    except ValidationError as ex:
        return ex.errors, ex.informal_only
    else:
        return [], True


time_spec = {
    'seconds': number(),
    'minutes': number(),
    'hours': number(),
    'days': number(),
    'weeks': number(),
    'time': anything(),
    'mtime': str(),
}

coverage = recursive({
    'polygons': str(),
    'polygons_srs': str(),
    'bbox': one_of(str(), [number()]),
    'bbox_srs': str(),
    'ogr_datasource': str(),
    'ogr_where': str(),
    'ogr_srs': str(),
    'datasource': one_of(str(), [number()]),
    'where': str(),
    'srs': str(),
    'expire_tiles': str(),
    'union': [recursive()],
    'difference': [recursive()],
    'intersection': [recursive()],
    'clip': bool(),
})

image_opts = {
    'mode': str(),
    'colors': number(),
    'transparent': bool(),
    'resampling_method': str(),
    'format': str(),
    'encoding_options': {
        anything(): anything()
    },
    'merge_method': str(),
}

http_opts = {
    'method': str(),
    'client_timeout': number(),
    'ssl_no_cert_checks': bool(),
    'ssl_ca_certs': str(),
    'hide_error_details': bool(),
    'headers': {
        anything(): str()
    },
    'manage_cookies': bool(),
}

mapserver_opts = {
    'binary': str(),
    'working_dir': str(),
}

scale_hints = {
    'max_scale': number(),
    'min_scale': number(),
    'max_res': number(),
    'min_res': number(),
}

source_commons = combined(
    scale_hints,
    {
        'concurrent_requests': int(),
        'coverage': coverage,
        'seed_only': bool(),
    }
)

cache_commons = combined(
    {
        'coverage': coverage,
    }
)

cache_types = {
    'file': combined(cache_commons, {
        'directory_layout': str(),
        'use_grid_names': bool(),
        'directory': str(),
        'tile_lock_dir': str(),
        'directory_permissions': str(),
        'file_permissions': str(),
    }),
    'sqlite': combined(cache_commons, {
        'directory': str(),
        'sqlite_timeout': number(),
        'sqlite_wal': bool(),
        'tile_lock_dir': str(),
        'ttl': int(),
        'directory_permissions': str(),
        'file_permissions': str(),
    }),
    'mbtiles': combined(cache_commons, {
        'filename': str(),
        'sqlite_timeout': number(),
        'sqlite_wal': bool(),
        'tile_lock_dir': str(),
        'directory_permissions': str(),
        'file_permissions': str(),
    }),
    'geopackage': combined(cache_commons, {
        'filename': str(),
        'directory': str(),
        'tile_lock_dir': str(),
        'table_name': str(),
        'levels': bool(),
        'directory_permissions': str(),
        'file_permissions': str(),
    }),
    'couchdb': combined(cache_commons, {
        'url': str(),
        'db_name': str(),
        'tile_metadata': {
            anything(): anything()
        },
        'tile_id': str(),
        'tile_lock_dir': str(),
    }),
    's3': combined(cache_commons, {
        'bucket_name': str(),
        'directory_layout': str(),
        'directory': str(),
        'profile_name': str(),
        'region_name': str(),
        'endpoint_url': str(),
        'access_control_list': str(),
        'tile_lock_dir': str(),
        'use_http_get': bool(),
        'include_grid_name': bool(),
    }),
    'redis': combined(cache_commons, {
        'host': str(),
        'port': int(),
        'password': str(),
        'username': str(),
        'db': int(),
        'prefix': str(),
        'default_ttl': int(),
        'ssl_certfile': str(),
        'ssl_keyfile': str(),
        'ssl_ca_certs': str(),
    }),
    'compact': combined(cache_commons, {
        'directory': str(),
        required('version'): number(),
        'tile_lock_dir': str(),
        'directory_permissions': str(),
        'file_permissions': str(),
    }),
    'azureblob': combined(cache_commons, {
        'connection_string': str(),
        'container_name': str(),
        'directory_layout': str(),
        'directory': str(),
        'tile_lock_dir': str(),
    }),
}

on_error = {
    anything(): {
        required('response'): one_of([int], str),
        'cache': bool,
        'authorize_stale': bool
    }
}


inspire_md = {
    'linked': {
        required('metadata_url'): {
            required('url'): str,
            required('media_type'): str,
        },
        required('languages'): {
            required('default'): str,
        },
    },
    'embedded': {
        required('resource_locators'): [{
            required('url'): str,
            required('media_type'): str,
        }],
        required('temporal_reference'): {
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
        },
        required('conformities'): [{
            'title': str,
            'uris': [str],
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
            required('resource_locators'): [{
                required('url'): str,
                required('media_type'): str,
            }],
            required('degree'): str,
        }],
        required('metadata_points_of_contact'): [{
            'organisation_name': str,
            'email': str,
        }],
        required('mandatory_keywords'): [str],
        'keywords': [{
            required('title'): str,
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
            'uris': [str],
            'resource_locators': [{
                required('url'): str,
                required('media_type'): str,
            }],
            required('keyword_value'): str,
        }],
        required('metadata_date'): one_of(str, datetime.date),
        'metadata_url': {
            required('url'): str,
            required('media_type'): str,
        },
        required('languages'): {
            required('default'): str,
        },
    },
}

wms_130_layer_md = {
    'abstract': str,
    'keyword_list': [
        {
            'vocabulary': str,
            'keywords': [str],
        }
    ],
    'attribution': {
        'title': str,
        'url':    str,
        'logo': {
            'url':    str,
            'width':  int,
            'height': int,
            'format': str,
        }
    },
    'identifier': [
        {
            'url': str,
            'name': str,
            'value': str,
        }
    ],
    'metadata': [
        {
            'url': str,
            'type': str,
            'format': str,
        },
    ],
    'data': [
        {
            'url': str,
            'format': str,
        }

    ],
    'feature_list': [
        {
            'url': str,
            'format': str,
        }
    ],
}

grid_opts = {
    'base': str(),
    'name': str(),
    'srs': str(),
    'bbox': one_of(str(), [number()]),
    'bbox_srs': str(),
    'num_levels': int(),
    'res': [number()],
    'res_factor': one_of(number(), str()),
    'max_res': number(),
    'min_res': number(),
    'stretch_factor': number(),
    'max_shrink_factor': number(),
    'align_resolutions_with': str(),
    'origin': str(),
    'tile_size': [int()],
    'threshold_res': [number()],
}

ogc_service_md = {
    'title': str,
    'abstract': str,
    'online_resource': str,
    'contact': anything(),
    'fees': str,
    'access_constraints': str,
    'keyword_list': [
        {
            'vocabulary': str,
            'keywords': [str],
        }
    ],
}

band_source = {
    required('source'): str(),
    required('band'): int,
    'factor': number(),
}

band_sources = {
    'r': [band_source],
    'g': [band_source],
    'b': [band_source],
    'a': [band_source],
    'l': [band_source],
}

mapproxy_yaml_spec = {
    '__config_files__': anything(),  # only used internaly
    'globals': {
        'image': {
            'resampling_method': 'method',
            'paletted': bool(),
            'stretch_factor': number(),
            'max_shrink_factor': number(),
            'jpeg_quality': number(),
            'formats': {
                anything(): image_opts,
            },
            'font_dir': str(),
            'merge_method': str(),
        },
        'http': combined(
            http_opts,
            {
                'access_control_allow_origin': one_of(str(), {}),
            }
        ),
        'cache': {
            'base_dir': str(),
            'lock_dir': str(),
            'tile_lock_dir': str(),
            'directory_permissions': str(),
            'file_permissions': str(),
            'meta_size': [number()],
            'meta_buffer': number(),
            'bulk_meta_tiles': bool(),
            'max_tile_limit': number(),
            'minimize_meta_requests': bool(),
            'concurrent_tile_creators': int(),
            'link_single_color_images': one_of(bool(), 'symlink', 'hardlink'),
            's3': {
                'bucket_name': str(),
                'profile_name': str(),
                'region_name': str(),
                'endpoint_url': str(),
            },
            'azureblob': {
                'connection_string': str(),
                'container_name': str(),
            },
        },
        'grid': {
            'tile_size': [int()],
        },
        'srs': {
            'axis_order_ne': [str()],
            'axis_order_en': [str()],
            'proj_data_dir': str(),
            'preferred_src_proj': {anything(): [str()]},
        },
        'tiles': {
            'expires_hours': number(),
        },
        'mapserver': mapserver_opts,
        'renderd': {
            'address': str(),
        }
    },
    'grids': {
        anything(): grid_opts,
    },
    'caches': {
        anything(): {
            required('sources'): one_of([str], band_sources),
            'name': str(),
            'grids': [str()],
            'cache_dir': str(),
            'meta_size': [number()],
            'meta_buffer': number(),
            'bulk_meta_tiles': bool(),
            'minimize_meta_requests': bool(),
            'concurrent_tile_creators': int(),
            'disable_storage': bool(),
            'format': str(),
            'image': image_opts,
            'request_format': str(),
            'use_direct_from_level': number(),
            'use_direct_from_res': number(),
            'link_single_color_images': one_of(bool(), 'symlink', 'hardlink'),
            'cache_rescaled_tiles': bool(),
            'upscale_tiles': int(),
            'downscale_tiles': int(),
            'refresh_before': time_spec,
            'watermark': {
                'text': str,
                'font_size': number(),
                'color': one_of(str(), [number()]),
                'opacity': number(),
                'spacing': str(),
            },
            'cache': type_spec('type', cache_types)
        }
    },
    'services': {
        'demo': {},
        'kml': {
            'use_grid_names': bool(),
        },
        'tms': {
            'use_grid_names': bool(),
            'origin': str(),
        },
        'wmts': {
            'kvp': bool(),
            'restful': bool(),
            'restful_template': str(),
            'restful_featureinfo_template': str(),
            'md': ogc_service_md,
            'featureinfo_formats': [
                {
                    required('mimetype'): str(),
                    'suffix': str(),
                },
            ],
        },
        'wms': {
            'srs': [str()],
            'bbox_srs': [one_of(str(), {'bbox': [number()], 'srs': str()})],
            'image_formats': [str()],
            'attribution': {
                'text': str,
            },
            'featureinfo_types': [str()],
            'featureinfo_xslt': {
                anything(): str()
            },
            'on_source_errors': str(),
            'max_output_pixels': one_of(number(), [number()]),
            'strict': bool(),
            'md': ogc_service_md,
            'inspire_md': type_spec('type', inspire_md),
            'versions': [str()],
        },
        'ogcapi': {
            'enable_tiles': bool(),
            'enable_maps': bool(),
            'map_srs': [str()],
            'attribution': {
                'text': str,
            },
            'on_source_errors': str(),
            'max_output_pixels': one_of(number(), [number()]),
            'image_formats': [str()],
            'concurrent_layer_renderer': number(),
            'md': {
                'identification': {
                    'title': str(),
                    'description': str(),
                    'url': str(),
                    'terms_of_service': str(),
                    'keywords': anything(),
                },
                'contact': {
                    'name': str(),
                    'email': str(),
                    'address': str(),
                    'city': str(),
                    'stateorprovince': str(),
                    'postalcode': str(),
                    'country': str(),
                    'phone': str(),
                    'fax': str(),
                    'url': str(),
                    'instructions': str(),
                },
                'license': {
                    required('name'): str(),
                    'url': str(),
                },
                'provider': {
                    'name': str(),
                    'url': str(),
                },
            },
            "default_dataset_layers": [str()],
        }
    },

    'sources': {
        anything(): type_spec('type', {
            'wms': combined(source_commons, {
                'wms_opts': {
                    'version': str(),
                    'map': bool(),
                    'featureinfo': bool(),
                    'legendgraphic': bool(),
                    'legendurl': str(),
                    'featureinfo_format': str(),
                    'featureinfo_xslt': str(),
                    'featureinfo_out_format': str(),
                    'query_layers': str(),
                },
                'image': combined(image_opts, {
                    'opacity': number(),
                    'transparent_color': one_of(str(), [number()]),
                    'transparent_color_tolerance': number(),
                }),
                'supported_formats': [str()],
                'supported_srs': [str()],
                'http': http_opts,
                'on_error': on_error,
                'forward_req_params': [str()],
                required('req'): {
                    required('url'): str(),
                    anything(): anything()
                }
            }),
            'mapserver': combined(source_commons, {
                'wms_opts': {
                    'version': str(),
                    'map': bool(),
                    'featureinfo': bool(),
                    'legendgraphic': bool(),
                    'legendurl': str(),
                    'featureinfo_format': str(),
                    'featureinfo_xslt': str(),
                },
                'image': combined(image_opts, {
                    'opacity': number(),
                    'transparent_color': one_of(str(), [number()]),
                    'transparent_color_tolerance': number(),
                }),
                'supported_formats': [str()],
                'supported_srs': [str()],
                'forward_req_params': [str()],
                required('req'): {
                    required('map'): str(),
                    anything(): anything()
                },
                'mapserver': mapserver_opts,
            }),
            'tile': combined(source_commons, {
                required('url'): str(),
                'transparent': bool(),
                'image': image_opts,
                'grid': str(),
                'request_format': str(),
                'origin': str(),  # TODO: remove with 1.5
                'http': http_opts,
                'on_error': on_error,
            }),
            'mapnik': combined(source_commons, {
                required('mapfile'): str(),
                'transparent': bool(),
                'image': image_opts,
                'layers': one_of(str(), [str()]),
                'use_mapnik2': bool(),
                'scale_factor': number(),
                'multithreaded': bool(),
            }),
            'arcgis': combined(source_commons, {
                required('req'): {
                    required('url'): str(),
                    'dpi': int(),
                    'layers': str(),
                    'transparent': bool(),
                    'time': str()
                },
                'opts': {
                    'featureinfo': bool(),
                    'featureinfo_tolerance': number(),
                    'featureinfo_return_geometries': bool(),
                },
                'supported_srs': [str()],
                'http': http_opts,
                'on_error': on_error
            }),
            'debug': {
            },
            'ogcapitiles': combined(source_commons, {
                required('landingpage_url'): str(),
                'collection': str(),
                'tile_matrix_set_id': str(),
                'image': image_opts,
            }),
            'ogcapimaps': combined(source_commons, {
                required('landingpage_url'): str(),
                'collection': str(),
                'supported_srs': [str()],
                'transparent': bool(),
                'image': combined(image_opts, {
                    'transparent_color': one_of(str(), [number()]),
                    'transparent_color_tolerance': number(),
                }),
                'bgcolor': one_of(str(), [number()]),
            }),
        })
    },

    'layers': one_of(
        {
            anything(): combined(scale_hints, {
                'sources': [str],
                required('title'): str,
                'legendurl': str(),
                'md': wms_130_layer_md,
            })
        },
        recursive([combined(scale_hints, {
            'sources': [str],
            'tile_sources': [str],
            'name': str(),
            required('title'): str,
            'legendurl': str(),
            'wmts_rest_legendurl': str(),
            'wmts_kvp_legendurl': str(),
            'layers': recursive(),
            'md': wms_130_layer_md,
            'dimensions': {
                anything(): {
                    required('values'): [one_of(str, float, int)],
                    'default': one_of(str, float, int),
                }
            },
            "nominal_res": number(),
            "nominal_scale": number(),
        })])
    ),
    # `parts` can be used for partial configurations that are referenced
    # from other sections (e.g. coverages, dimensions, etc.)
    'parts': anything(),
}


def add_source_to_mapproxy_yaml_spec(source_name, source_spec):
    """ Add a new source type to mapproxy_yaml_spec.
        Used by plugins.
    """

    # sources has a single anything() : {} member
    values = list(mapproxy_yaml_spec['sources'].values())
    assert len(values) == 1
    values[0].add_subspec(source_name, source_spec)


def add_service_to_mapproxy_yaml_spec(service_name, service_spec):
    """ Add a new service type to mapproxy_yaml_spec.
        Used by plugins.
    """

    mapproxy_yaml_spec['services'][service_name] = service_spec


def add_subcategory_to_layer_md(category_name, category_def):
    """ Add a new category to wms_130_layer_md.
        Used by plugins
    """
    wms_130_layer_md[category_name] = category_def



# -*- coding: utf-8 -*-
"""
This module implements context-local objects.

This is a partial version of werkzeug/local.py containing only Local and
StackLocal.

Last update: 2011-03-15 9ada59c958b2edbb9739fb55a6b32ef4a97dac07

:copyright: (c) 2010 by the Werkzeug Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
from werkzeug.local import LocalProxy

try:
    from greenlet import getcurrent as get_current_greenlet
except ImportError:  # pragma: no cover
    try:
        from py.magic import greenlet
        get_current_greenlet = greenlet.getcurrent
        del greenlet
    except Exception:
        # catch all, py.* fails with so many different errors.
        get_current_greenlet = int
try:
    from _thread import get_ident as get_current_thread, allocate_lock
except ImportError:  # pragma: no cover
    try:
        from thread import get_ident as get_current_thread, allocate_lock
    except ImportError:  # pragma: no cover
        from dummy_thread import get_ident as get_current_thread, allocate_lock


# get the best ident function.  if greenlets are not installed we can
# safely just use the builtin thread function and save a python methodcall
# and the cost of calculating a hash.
if get_current_greenlet is int:  # pragma: no cover
    get_ident = get_current_thread
else:
    def get_ident(): return (get_current_thread(), get_current_greenlet())  # noqa


def release_local(local):
    """Releases the contents of the local for the current context.
    This makes it possible to use locals without a manager.

    Example::

        >>> loc = Local()
        >>> loc.foo = 42
        >>> release_local(loc)
        >>> hasattr(loc, 'foo')
        False

    With this function one can release :class:`Local` objects as well
    as :class:`StackLocal` objects.  However it is not possible to
    release data held by proxies that way, one always has to retain
    a reference to the underlying local object in order to be able
    to release it.

    .. versionadded:: 0.6.1
    """
    local.__release_local__()


class Local(object):
    __slots__ = ('__storage__', '__lock__', '__ident_func__')

    def __init__(self):
        object.__setattr__(self, '__storage__', {})
        object.__setattr__(self, '__lock__', allocate_lock())
        object.__setattr__(self, '__ident_func__', get_ident)

    def __iter__(self):
        return self.__storage__.iteritems()

    def __call__(self, proxy):
        """Create a proxy for a name."""
        return LocalProxy(self, proxy)

    def __release_local__(self):
        self.__storage__.pop(self.__ident_func__(), None)

    def __getattr__(self, name):
        try:
            return self.__storage__[self.__ident_func__()][name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        ident = self.__ident_func__()
        self.__lock__.acquire()
        try:
            storage = self.__storage__
            if ident in storage:
                storage[ident][name] = value
            else:
                storage[ident] = {name: value}
        finally:
            self.__lock__.release()

    def __delattr__(self, name):
        try:
            del self.__storage__[self.__ident_func__()][name]
        except KeyError:
            raise AttributeError(name)


class LocalStack(object):
    """This class works similar to a :class:`Local` but keeps a stack
    of objects instead.  This is best explained with an example::

        >>> ls = LocalStack()
        >>> ls.push(42)
        [42]
        >>> ls.top
        42
        >>> ls.push(23)
        [42, 23]
        >>> ls.top
        23
        >>> ls.pop()
        23
        >>> ls.top
        42

    They can be force released by using a :class:`LocalManager` or with
    the :func:`release_local` function but the correct way is to pop the
    item from the stack after using.  When the stack is empty it will
    no longer be bound to the current context (and as such released).

    By calling the stack without arguments it returns a proxy that resolves to
    the topmost item on the stack.

    .. versionadded:: 0.6.1
    """

    def __init__(self):
        self._local = Local()
        self._lock = allocate_lock()

    def __release_local__(self):
        self._local.__release_local__()

    def _get__ident_func__(self):
        return self._local.__ident_func__

    def _set__ident_func__(self, value):
        object.__setattr__(self._local, '__ident_func__', value)
    __ident_func__ = property(_get__ident_func__, _set__ident_func__)
    del _get__ident_func__, _set__ident_func__

    def __call__(self):
        def _lookup():
            rv = self.top
            if rv is None:
                raise RuntimeError('object unbound')
            return rv
        return LocalProxy(_lookup)

    def push(self, obj):
        """Pushes a new item to the stack"""
        self._lock.acquire()
        try:
            rv = getattr(self._local, 'stack', None)
            if rv is None:
                self._local.stack = rv = []
            rv.append(obj)
            return rv
        finally:
            self._lock.release()

    def pop(self):
        """Removes the topmost item from the stack, will return the
        old value or `None` if the stack was already empty.
        """
        self._lock.acquire()
        try:
            stack = getattr(self._local, 'stack', None)
            if stack is None:
                return None
            elif len(stack) == 1:
                release_local(self._local)
                return stack[-1]
            else:
                return stack.pop()
        finally:
            self._lock.release()

    @property
    def top(self):
        """The topmost item on the stack.  If the stack is empty,
        `None` is returned.
        """
        try:
            return self._local.stack[-1]
        except (AttributeError, IndexError):
            return None



# This file is part of the MapProxy project.
# Copyright (C) 2016 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import calendar
import hashlib
import sys
import threading

from mapproxy.image import ImageSource
from mapproxy.cache import path
from mapproxy.cache.base import tile_buffer, TileCacheBase
from mapproxy.util import async_
from mapproxy.util.py import reraise_exception
from urllib import request as urllib2

try:
    import boto3
    import botocore
except ImportError:
    boto3 = None


import logging
log = logging.getLogger('mapproxy.cache.s3')


_s3_sessions_cache = threading.local()


def s3_session(profile_name=None):
    if not hasattr(_s3_sessions_cache, 'sessions'):
        _s3_sessions_cache.sessions = {}
    if profile_name not in _s3_sessions_cache.sessions:
        _s3_sessions_cache.sessions[profile_name] = boto3.session.Session(profile_name=profile_name)
    return _s3_sessions_cache.sessions[profile_name]


class S3ConnectionError(Exception):
    pass


class S3Cache(TileCacheBase):

    def __init__(self, base_path, file_ext, directory_layout='tms',
                 bucket_name='mapproxy', profile_name=None, region_name=None, endpoint_url=None,
                 _concurrent_writer=4, access_control_list=None, coverage=None, use_http_get=False):
        super(S3Cache, self).__init__(coverage)
        md5 = hashlib.new('md5', base_path.encode('utf-8') + bucket_name.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = md5.hexdigest()
        self.bucket_name = bucket_name
        self.profile_name = profile_name
        self.region_name = region_name
        self.endpoint_url = endpoint_url
        self.access_control_list = access_control_list
        self.use_http_get = use_http_get

        try:
            self.bucket = self.conn().head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                raise S3ConnectionError('No such bucket: %s' % bucket_name)
            elif e.response['Error']['Code'] == '403':
                raise S3ConnectionError('Access denied. Check your credentials')
            else:
                reraise_exception(
                    S3ConnectionError('Unknown error: %s' % e),
                    sys.exc_info(),
                )

        self.base_path = base_path
        self.file_ext = file_ext
        self._concurrent_writer = _concurrent_writer

        self._tile_location, _ = path.location_funcs(layout=directory_layout)

    def get_bucket_url(self, tile):
        return f"https://{self.bucket_name}.s3.{self.region_name}.amazonaws.com/{self.tile_key(tile)}"

    def tile_key(self, tile):
        return self._tile_location(tile, self.base_path, self.file_ext).lstrip('/')

    def conn(self):
        if boto3 is None:
            raise ImportError("S3 Cache requires 'boto3' package.")

        try:
            return s3_session(self.profile_name).client(
                "s3", region_name=self.region_name, endpoint_url=self.endpoint_url)
        except Exception as e:
            raise S3ConnectionError('Error during connection %s' % e)

    def load_tile_metadata(self, tile, dimensions=None):
        if tile.timestamp:
            return
        self.is_cached(tile, dimensions=dimensions)

    def _set_metadata(self, response, tile):
        if 'LastModified' in response:
            tile.timestamp = calendar.timegm(response['LastModified'].timetuple())
        if 'ContentLength' in response:
            tile.size = response['ContentLength']

    def is_cached(self, tile, dimensions=None):
        if tile.is_missing():
            if self.use_http_get:
                try:
                    req = urllib2.Request(self.get_bucket_url(tile))
                    response = urllib2.urlopen(req)
                    self._set_metadata(response.info(), tile)
                except urllib2.HTTPError as e:
                    if e.code == 403:
                        return False
                    raise
            else:
                key = self.tile_key(tile)
                try:
                    r = self.conn().head_object(Bucket=self.bucket_name, Key=key)
                    self._set_metadata(r, tile)
                except botocore.exceptions.ClientError as e:
                    if e.response['Error']['Code'] in ('404', 'NoSuchKey'):
                        return False
                    raise

        return True

    def load_tiles(self, tiles, with_metadata=True, dimensions=None):
        p = async_.Pool(min(4, len(tiles)))
        return all(p.map(self.load_tile, tiles))

    def load_tile(self, tile, with_metadata=True, dimensions=None):
        if not tile.is_missing():
            return True

        key = self.tile_key(tile)
        log.debug('S3:load_tile, key: %s' % key)

        if self.use_http_get:
            try:
                req = urllib2.Request(self.get_bucket_url(tile))
                tile.source = ImageSource(urllib2.urlopen(req))
            except urllib2.HTTPError as e:
                if e.code == 403:
                    return False
                raise
        else:
            try:
                r = self.conn().get_object(Bucket=self.bucket_name, Key=key)
                self._set_metadata(r, tile)
                tile.source = ImageSource(r['Body'])
            except botocore.exceptions.ClientError as e:
                # moto get_object can return Error wrapped in Errors...
                error = e.response.get('Errors', e.response)['Error']
                if error['Code'] in ('404', 'NoSuchKey'):
                    return False
                raise

        return True

    def remove_tile(self, tile, dimensions=None):
        key = self.tile_key(tile)
        log.debug('remove_tile, key: %s' % key)
        self.conn().delete_object(Bucket=self.bucket_name, Key=key)

    def store_tiles(self, tiles, dimensions=None):
        p = async_.Pool(min(self._concurrent_writer, len(tiles)))
        p.map(self.store_tile, tiles)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return

        key = self.tile_key(tile)
        log.debug('S3: store_tile, key: %s' % key)

        extra_args = {}
        if self.file_ext in ('jpeg', 'png'):
            extra_args['ContentType'] = 'image/' + self.file_ext
        if self.access_control_list:
            extra_args['ACL'] = self.access_control_list
        with tile_buffer(tile) as buf:
            self.conn().upload_fileobj(
                NopCloser(buf),  # upload_fileobj closes buf, wrap in NopCloser
                self.bucket_name,
                key,
                ExtraArgs=extra_args)


class NopCloser(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def close(self):
        pass

    def __getattr__(self, name):
        return getattr(self.wrapped, name)



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from threading import Lock

from mapproxy.client.http import HTTPClientError
from mapproxy.grid.tile_grid import tile_grid_from_ogc_tile_matrix_set
from mapproxy.image.opts import ImageOptions
from mapproxy.layer import (
    BlankImage,
    MapExtent,
    DefaultMapExtent,
    MapLayer,
)
from mapproxy.source import SourceError, InvalidSourceQuery
from mapproxy.srs import ogc_crs_url_to_auth_code
from mapproxy.util.py import reraise_exception
from mapproxy.util.ogcapi import (
    find_href_in_links,
    normalize_srs_code,
    build_absolute_url,
)

import logging

log = logging.getLogger("mapproxy.source.ogcapitiles")
log_config = logging.getLogger("mapproxy.config")

# For testing
reset_config_cache = False


class OGCAPITilesSource(MapLayer):
    def __init__(
        self,
        landingpage_url,
        collection,
        http_client,
        tile_matrix_set_id=None,
        coverage=None,
        image_opts=None,
        error_handler=None,
        res_range=None,
    ):
        MapLayer.__init__(self, image_opts=image_opts)
        self.landingpage_url = landingpage_url.rstrip("/")
        self.collection = collection
        self.http_client = http_client
        self.image_opts = image_opts or ImageOptions()
        self.tile_matrix_set_id = tile_matrix_set_id
        self.coverage = coverage
        if self.coverage:
            self.extent = MapExtent(self.coverage.bbox, self.coverage.srs)
        else:
            self.extent = DefaultMapExtent()
        self.res_range = res_range
        self.error_handler = error_handler

        self.lock = Lock()
        # Below variables are protected under the lock
        self._reset_config_caches()

    def _reset_config_caches(self):
        self.has_got_tileset_list = False
        self.map_crs_to_tilesets_list = {}
        self.map_srs_to_grid_and_template_url = {}
        self.map_grid_name_to_cachemaplayer = {}

    def _build_url(self, href):
        return build_absolute_url(self.landingpage_url, href)

    def _get_tileset_list(self):
        with self.lock:
            # Used by tests to avoid caching of "metadata"
            global reset_config_cache
            if reset_config_cache:
                reset_config_cache = False
                self._reset_config_caches()

            if self.has_got_tileset_list:
                return

            headers = {"Accept": "application/json"}

            url = self.landingpage_url
            if self.collection:
                url += "/collections/" + self.collection
            try:
                resp = self.http_client.open(url, headers=headers)
            except HTTPClientError as e:
                log.warning(f"Cannot retrieve {url}: %s", e)
                reraise_exception(SourceError(e.args[0]), sys.exc_info())
            try:
                j = json.loads(resp.read().decode("utf-8"))
            except Exception as e:
                log.warning(f"Cannot parse response to {url} as JSON: %s", e)
                reraise_exception(SourceError(e.args[0]), sys.exc_info())

            if "links" not in j:
                ex = SourceError(f"Could not retrieve 'links' in {url} response")
                log.error(ex)
                raise ex

            tilesets_map_href = find_href_in_links(
                j["links"],
                "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                "application/json",
            )
            if not tilesets_map_href:
                ex = SourceError(
                    f"Could not retrieve a tilesets-map link in {url} response"
                )
                log.error(ex)
                raise ex

            tilesets_map_url = self._build_url(tilesets_map_href)
            try:
                resp = self.http_client.open(tilesets_map_url, headers=headers)
            except HTTPClientError as e:
                log.warning(f"Cannot retrieve {tilesets_map_url}: %s", e)
                reraise_exception(SourceError(e.args[0]), sys.exc_info())
            try:
                j = json.loads(resp.read().decode("utf-8"))
            except Exception as e:
                log.warning(
                    f"Cannot parse response to {tilesets_map_url} as JSON: %s", e
                )
                reraise_exception(SourceError(e.args[0]), sys.exc_info())

            if "tilesets" not in j:
                ex = SourceError(
                    f"Could not retrieve 'tilesets' in {tilesets_map_url} response"
                )
                log.error(ex)
                raise ex

            map_tilesets_candidates = filter(
                lambda t: t["dataType"] == "map", j["tilesets"]
            )

            if self.tile_matrix_set_id:
                user_matrix_sets = []
                for tileset in map_tilesets_candidates:
                    tileset_links = tileset["links"]
                    for link in tileset_links:
                        if "href" in link and link["href"].split("?")[0].endswith(
                            "/" + self.tile_matrix_set_id
                        ):
                            user_matrix_sets.append(tileset)
                            break
                map_tilesets_candidates = user_matrix_sets

            for tileset in map_tilesets_candidates:
                crs = normalize_srs_code(ogc_crs_url_to_auth_code(tileset["crs"]))
                if crs not in self.map_crs_to_tilesets_list:
                    self.map_crs_to_tilesets_list[crs] = []
                self.map_crs_to_tilesets_list[crs].append(tileset)

            self.has_got_tileset_list = True

    def _get_grid_and_template_url_from_tileset(self, tileset, image_mime_type):
        links = tileset["links"]
        tiling_scheme_href = find_href_in_links(
            links,
            "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
            "application/json",
        )
        if not tiling_scheme_href:
            ex = SourceError(
                f"Could not retrieve a 'tiling-scheme' link for tileset {tileset}"
            )
            log.error(ex)
            raise ex

        tiling_scheme_url = self._build_url(tiling_scheme_href)

        tileset_href = find_href_in_links(links, "self", "application/json")
        if not tiling_scheme_href:
            ex = SourceError(f"Could not retrieve a 'self' link for tileset {tileset}")
            log.error(ex)
            raise ex

        tileset_url = self._build_url(tileset_href)

        headers = {"Accept": "application/json"}
        try:
            resp = self.http_client.open(tiling_scheme_url, headers=headers)
        except HTTPClientError as e:
            log.warning(f"Cannot retrieve {tiling_scheme_url}: %s", e)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())
        try:
            tile_matrix_set = json.loads(resp.read().decode("utf-8"))
        except Exception as e:
            log.warning(f"Cannot parse response to {tiling_scheme_url} as JSON: %s", e)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())
        grid = tile_grid_from_ogc_tile_matrix_set(tile_matrix_set)

        try:
            resp = self.http_client.open(tileset_url, headers=headers)
        except HTTPClientError as e:
            log.warning(f"Cannot retrieve {tileset_url}: %s", e)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())
        try:
            tileset_full = json.loads(resp.read().decode("utf-8"))
        except Exception as e:
            log.warning(f"Cannot parse response to {tileset_url} as JSON: %s", e)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())

        template_href = find_href_in_links(
            tileset_full["links"], "item", image_mime_type
        )
        if not template_href:
            ex = SourceError(
                f"Could not retrieve a tile template URL for tileset {tileset}"
            )
            log.error(ex)
            raise ex

        return grid, self._build_url(template_href)

    def _get_grid_and_template_url_from_srs(self, query_srs, image_mime_type):
        srs_code = normalize_srs_code(query_srs.srs_code)
        key = srs_code + "/" + image_mime_type

        with self.lock:
            grid_and_template_url = self.map_srs_to_grid_and_template_url.get(key, None)
            if grid_and_template_url:
                return grid_and_template_url

            if srs_code in self.map_crs_to_tilesets_list:
                for tileset in self.map_crs_to_tilesets_list[srs_code]:
                    try:
                        grid_and_template_url = (
                            self._get_grid_and_template_url_from_tileset(
                                tileset, image_mime_type
                            )
                        )
                        break
                    except Exception as e:
                        log.info(f"Exception while evaluating tileset {tileset}: {e}")
                        pass

            if grid_and_template_url is None:
                ex = SourceError(f"Cannot find a valid tile matrix set for {query_srs}")
                log.error(ex)
                raise ex

            self.map_srs_to_grid_and_template_url[key] = grid_and_template_url
            return grid_and_template_url

    def get_map(self, query):
        self._get_tileset_list()
        image_mime_type = "image/" + query.format
        grid, template_url = self._get_grid_and_template_url_from_srs(
            query.srs, image_mime_type
        )

        if grid.tile_size != query.size:
            ex = InvalidSourceQuery(
                "tile size of cache and tile source do not match: %s != %s"
                % (grid.tile_size, query.size)
            )
            log_config.error(ex)
            raise ex

        if grid.srs != query.srs:
            ex = InvalidSourceQuery(
                "SRS of cache and tile source do not match: %r != %r"
                % (grid.srs, query.srs)
            )
            log_config.error(ex)
            raise ex

        if self.res_range and not self.res_range.contains(
            query.bbox, query.size, query.srs
        ):
            raise BlankImage()
        if self.coverage and not self.coverage.intersects(query.bbox, query.srs):
            raise BlankImage()

        _bbox, grid, tiles = grid.get_affected_tiles(query.bbox, query.size)

        if grid != (1, 1):
            raise InvalidSourceQuery("BBOX does not align to tile")

        x, y, z = next(tiles)
        tile_url = (
            template_url.replace("{tileMatrix}", str(z))
            .replace("{tileRow}", str(y))
            .replace("{tileCol}", str(x))
        )
        try:
            return self.http_client.open_image(tile_url)
        except HTTPClientError as e:
            if self.error_handler:
                resp = self.error_handler.handle(e.response_code, query)
                if resp:
                    return resp
            log.warning("could not retrieve tile: %s", e)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())



# This file is part of the MapProxy project.
# Copyright (C) 2010-2017 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tile retrieval (WMS, TMS, etc.).
"""
import sys
import time
from typing import Any

from mapproxy.version import version
from mapproxy.image import ImageSource
from mapproxy.util.py import reraise_exception
from mapproxy.client.log import log_request

from urllib import request as urllib2
from urllib.error import URLError, HTTPError
from urllib.request import HTTPCookieProcessor
from http import client as httplib
from http.cookiejar import CookieJar

import socket
import ssl


class HTTPClientError(Exception):
    def __init__(self, arg, response_code=None, full_msg=None):
        Exception.__init__(self, arg)
        self.response_code = response_code
        self.full_msg = full_msg


def build_https_handler(ssl_ca_certs, insecure):
    if insecure:
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
    elif ssl_ca_certs:
        ctx = ssl.create_default_context(cafile=ssl_ca_certs)
    else:
        ctx = ssl.create_default_context()
    return urllib2.HTTPSHandler(context=ctx)


class VerifiedHTTPSConnection(httplib.HTTPSConnection):
    def __init__(self, *args, **kw):
        self._ca_certs = kw.pop('ca_certs', None)
        httplib.HTTPSConnection.__init__(self, *args, **kw)

    def connect(self):
        # overrides the version in httplib so that we do
        #    certificate verification

        if hasattr(socket, 'create_connection') and hasattr(self, 'source_address'):
            sock = socket.create_connection((self.host, self.port),
                                            self.timeout, self.source_address)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))

        if hasattr(self, '_tunnel_host') and self._tunnel_host:
            # for Python >= 2.6 with proxy support
            self.sock = sock
            self._tunnel()

        # wrap the socket using verification with the root
        #    certs in self.ca_certs_path
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_cert_chain(certfile=self.cert_file, keyfile=self.key_file)
        context.load_verify_locations(cafile=self._ca_certs)
        context.minimum_version = ssl.TLSVersion.TLSv1_2

        self.sock = context.wrap_socket(sock, server_hostname=self.host)


def verified_https_connection_with_ca_certs(ca_certs):
    """
    Creates VerifiedHTTPSConnection classes with given ca_certs file.
    """
    def wrapper(*args, **kw):
        kw['ca_certs'] = ca_certs
        return VerifiedHTTPSConnection(*args, **kw)
    return wrapper


class VerifiedHTTPSHandler(urllib2.HTTPSHandler):
    def __init__(self, connection_class=VerifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)

    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)


class _URLOpenerCache(object):
    """
    Creates custom URLOpener with BasicAuth and HTTPS handler.

    Caches and reuses opener if possible (i.e. if they share the same
    ssl_ca_certs).
    """

    def __init__(self):
        self._opener = {}

    def __call__(self, ssl_ca_certs, url, username, password, insecure=False, manage_cookies=False):
        cache_key = (ssl_ca_certs, insecure, manage_cookies)
        if cache_key not in self._opener:
            handlers = []
            https_handler = build_https_handler(ssl_ca_certs, insecure)
            if https_handler:
                handlers.append(https_handler)
            passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
            authhandler = urllib2.HTTPBasicAuthHandler(passman)
            handlers.append(authhandler)
            authhandler = urllib2.HTTPDigestAuthHandler(passman)
            handlers.append(authhandler)
            if manage_cookies:
                cj = CookieJar()
                handlers.append(HTTPCookieProcessor(cj))

            opener = urllib2.build_opener(*handlers)

            opener.addheaders = [('User-agent', 'MapProxy-%s' % (version,))]

            self._opener[cache_key] = (opener, passman)
        else:
            opener, passman = self._opener[cache_key]

        if url is not None and username is not None and password is not None:
            passman.add_password(None, url, username, password)

        return opener


create_url_opener = _URLOpenerCache()


class HTTPClient(object):
    def __init__(self, url=None, username=None, password=None, insecure=False,
                 ssl_ca_certs=None, timeout=None, headers=None, hide_error_details=False,
                 manage_cookies=False):
        self._timeout = timeout
        if url and url.startswith('https') and insecure:
            ssl_ca_certs = None

        self.opener = create_url_opener(ssl_ca_certs, url, username, password,
                                        insecure=insecure, manage_cookies=manage_cookies)
        self.headers = headers if headers else {}
        self.hide_error_details = hide_error_details

    def open(self, url, data=None, method=None, headers=None) -> Any:

        """
        Open a url and return a HTTP response.

        Parameters
        ----------
        url : string
            URL
        data : object | None
            additional data to send to the server.
        method : string | None
             HTTP request method that will be used.
        headers : dict | None
            Dictionary with key:value pairs to pass as headers. This is
            merged with the headers potentially passed to the constructor of
            HTTPClient.

        Returns
        -------
        A slightly modified http.client.HTTPResponse object.
        """

        code = None
        result = None
        start_time = time.time()
        req = None
        try:
            req = urllib2.Request(url, data=data)
        except ValueError as e:
            err = self.handle_url_exception(url, 'URL not correct', e.args[0])
            reraise_exception(err, sys.exc_info())
        headers = self.headers | (headers if headers else {})
        for key, value in headers.items():
            req.add_header(key, value)
        if method:
            req.method = method
        try:
            if self._timeout is not None:
                result = self.opener.open(req, timeout=self._timeout)
            else:
                result = self.opener.open(req)
        except HTTPError as e:
            code = e.code
            err = self.handle_url_exception(url, 'HTTP Error', str(code), response_code=code)
            reraise_exception(err, sys.exc_info())
        except URLError as e:
            if isinstance(e.reason, ssl.SSLError):
                err = self.handle_url_exception(url, 'Could not verify connection to URL', e.reason.args[1])
                reraise_exception(err, sys.exc_info())
            try:
                reason = e.reason.args[1]
            except (AttributeError, IndexError):
                reason = e.reason
            err = self.handle_url_exception(url, 'No response from URL', reason)
            reraise_exception(err, sys.exc_info())
        except ValueError as e:
            err = self.handle_url_exception(url, 'URL not correct', e.args[0])
            reraise_exception(err, sys.exc_info())
        except Exception as e:
            err = self.handle_url_exception(url, 'Internal HTTP error', repr(e))
            reraise_exception(err, sys.exc_info())
        else:
            code = getattr(result, 'code', 200)
            if code == 204:
                raise HTTPClientError('HTTP Error "204 No Content"', response_code=204)
            return result
        finally:
            log_request(url, code, result, duration=time.time()-start_time, method=req.get_method())

    def open_image(self, url, data=None):
        resp = self.open(url, data=data)
        if 'content-type' in resp.headers:
            if not resp.headers['content-type'].lower().startswith('image'):
                raise HTTPClientError('response is not an image: (%s)' % (resp.read()))
        return ImageSource(resp)

    def handle_url_exception(self, url, message, reason, response_code=None):
        full_msg = '%s "%s": %s' % (message, url, reason)
        if self.hide_error_details:
            return HTTPClientError(
                '{} (see logs for URL and reason).'.format(message),
                response_code=response_code,
                full_msg=full_msg,
            )
        else:
            return HTTPClientError(
                full_msg,
                response_code=response_code,
            )


def auth_data_from_url(url):
    """
    >>> auth_data_from_url('invalid_url')
    ('invalid_url', (None, None))
    >>> auth_data_from_url('http://localhost/bar')
    ('http://localhost/bar', (None, None))
    >>> auth_data_from_url('http://bar@localhost/bar')
    ('http://localhost/bar', ('bar', None))
    >>> auth_data_from_url('http://bar:baz@localhost/bar')
    ('http://localhost/bar', ('bar', 'baz'))
    >>> auth_data_from_url('http://bar:b:az@@localhost/bar')
    ('http://localhost/bar', ('bar', 'b:az@'))
    >>> auth_data_from_url('http://bar foo; foo@bar:b:az@@localhost/bar')
    ('http://localhost/bar', ('bar foo; foo@bar', 'b:az@'))
    >>> auth_data_from_url('https://bar:foo#;%$@localhost/bar')
    ('https://localhost/bar', ('bar', 'foo#;%$'))
    >>> auth_data_from_url('http://localhost/bar@2x')
    ('http://localhost/bar@2x', (None, None))
    >>> auth_data_from_url('http://bar@localhost/bar@2x')
    ('http://localhost/bar@2x', ('bar', None))
    >>> auth_data_from_url('http://bar:baz@localhost/bar@2x')
    ('http://localhost/bar@2x', ('bar', 'baz'))
    >>> auth_data_from_url('https://bar@localhost/bar/0/0/0@2x.png')
    ('https://localhost/bar/0/0/0@2x.png', ('bar', None))
    >>> auth_data_from_url('http://bar:baz@localhost/bar@2x.png')
    ('http://localhost/bar@2x.png', ('bar', 'baz'))
    """
    if not url or '://' not in url:
        # be silent for invalid URLs
        return url, (None, None)

    schema, url = url.split('://', 1)
    if '/' in url:
        host, request = url.split('/', 1)
    else:
        host, request = url, ''

    username = password = None
    if '@' in host:
        auth_data, host = host.rsplit('@', 1)
        if ':' in auth_data:
            username, password = auth_data.split(':', 1)
        else:
            username = auth_data
    url = schema + "://" + host + "/" + request
    return url, (username, password)


def open_url(url):
    url, (username, password) = auth_data_from_url(url)
    http_client = HTTPClient(url, username, password)
    return http_client.open(url)


def retrieve_image(url, client=None):
    """
    Retrive an image from `url`.

    :return: the image as a file object (with url .header and .info)
    :raise HTTPClientError: if response content-type doesn't start with image
    """
    resp = open_url(url)
    if not resp.headers['content-type'].startswith('image'):
        raise HTTPClientError('response is not an image: (%s)' % (resp.read()))
    return ImageSource(resp)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Service responses.
"""

import copy
import hashlib
from mapproxy.util.times import format_httpdate, parse_httpdate, timestamp


class Response(object):
    charset = 'utf-8'
    default_content_type = 'text/plain'
    block_size = 1024 * 32

    def __init__(self, response, status=None, content_type=None, mimetype=None, headers=None):
        self.response = response
        if status is None:
            status = 200
        self.status = status
        self._timestamp = None
        self.headers = copy.copy(headers) if headers else {}
        if mimetype:
            if mimetype.startswith('text/'):
                content_type = mimetype + '; charset=' + self.charset
            else:
                content_type = mimetype
        if content_type is None:
            content_type = self.default_content_type
        self.headers['Content-type'] = content_type

        if content_type.startswith(('text/', 'application/')):
            # Capability documents can be dependent on the value of a few X-headers.
            # Tell this caching proxies via the Vary HTTP header. This also prevents
            # malicious cache poisoning.
            self.headers['Vary'] = 'X-Script-Name, X-Forwarded-Host, X-Forwarded-Proto'

    def _status_set(self, status):
        if isinstance(status, int):
            status = status_code(status)
        self._status = status

    def _status_get(self):
        return self._status

    status = property(_status_get, _status_set)

    def _last_modified_set(self, date):
        if not date:
            return
        self._timestamp = timestamp(date)
        self.headers['Last-modified'] = format_httpdate(self._timestamp)

    def _last_modified_get(self):
        return self.headers.get('Last-modified', None)

    last_modified = property(_last_modified_get, _last_modified_set)

    def _etag_set(self, value):
        self.headers['ETag'] = value

    def _etag_get(self):
        return self.headers.get('ETag', None)

    etag = property(_etag_get, _etag_set)

    def cache_headers(self, timestamp=None, etag_data=None, max_age=None, no_cache=False):
        """
        Set cache-related headers.

        :param timestamp: local timestamp of the last modification of the
            response content
        :param etag_data: list that will be used to build an ETag hash.
            calls the str function on each item.
        :param max_age: the maximum cache age in seconds
        """
        if etag_data:
            hash_src = ''.join((str(x) for x in etag_data)).encode('ascii')
            self.etag = hashlib.new('md5', hash_src, usedforsecurity=False).hexdigest()

        if no_cache:
            assert not timestamp and not max_age
            self.headers['Cache-Control'] = 'no-cache, no-store'
            self.headers['Pragma'] = 'no-cache'
            self.headers['Expires'] = '-1'

        self.last_modified = timestamp
        if (timestamp or etag_data) and max_age is not None:
            self.headers['Cache-control'] = 'public, max-age=%d, s-maxage=%d' % (max_age, max_age)

    def make_conditional(self, req):
        """
        Make the response conditional to the HTTP headers in the CGI/WSGI `environ`.
        Checks for ``If-none-match`` and ``If-modified-since`` headers and compares
        to the etag and timestamp of this response. If the content was not modified
        the repsonse will changed to HTTP 304 Not Modified.
        """
        if req is None:
            return
        environ = req.environ

        not_modified = False

        if self.etag == environ.get('HTTP_IF_NONE_MATCH', -1):
            not_modified = True
        elif self._timestamp is not None:
            date = environ.get('HTTP_IF_MODIFIED_SINCE', None)
            timestamp = parse_httpdate(date)
            if timestamp is not None and self._timestamp <= timestamp:
                not_modified = True

        if not_modified:
            self.status = 304
            self.response = []
            if 'Content-type' in self.headers:
                del self.headers['Content-type']

    @property
    def content_length(self):
        return int(self.headers.get('Content-length', 0))

    @property
    def content_type(self):
        return self.headers['Content-type']

    @property
    def data(self):
        if hasattr(self.response, 'read'):
            return self.response.read()
        else:
            return b''.join(chunk.encode() for chunk in self.response)

    @property
    def fixed_headers(self):
        headers = []
        for key, value in self.headers.items():
            if type(value) is not str:
                # for str subclasses like ImageFormat
                value = str(value)
            headers.append((key, value))
        return headers

    def __call__(self, environ, start_response):
        if hasattr(self.response, 'read'):
            if ((not hasattr(self.response, 'ok_to_seek') or
                self.response.ok_to_seek) and
               (hasattr(self.response, 'seek') and
                    hasattr(self.response, 'tell'))):
                self.response.seek(0, 2)  # to EOF
                self.headers['Content-length'] = str(self.response.tell())
                self.response.seek(0)
            if 'wsgi.file_wrapper' in environ:
                resp_iter = environ['wsgi.file_wrapper'](self.response, self.block_size)
            else:
                resp_iter = iter(lambda: self.response.read(self.block_size), b'')
        elif not self.response:
            resp_iter = iter([])
        elif isinstance(self.response, str):
            self.response = self.response.encode(self.charset)
            self.headers['Content-length'] = str(len(self.response))
            resp_iter = iter([self.response])
        elif isinstance(self.response, bytes):
            self.headers['Content-length'] = str(len(self.response))
            resp_iter = iter([self.response])
        else:
            resp_iter = self.response

        start_response(self.status, self.fixed_headers)
        return resp_iter

    def iter_encode(self, chunks):
        for chunk in chunks:
            if isinstance(chunk, str):
                chunk = chunk.encode(self.charset)
            yield chunk


# http://www.faqs.org/rfcs/rfc2616.html
_status_codes = {
    100: 'Continue',
    101: 'Switching Protocols',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Time-out',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Large',
    415: 'Unsupported Media Type',
    416: 'Requested range not satisfiable',
    417: 'Expectation Failed',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Time-out',
    505: 'HTTP Version not supported',
}


def status_code(code):
    return str(code) + ' ' + _status_codes[code]



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from mapproxy.client.http import HTTPClient
from mapproxy.script.wms_capabilities import wms_capabilities_command
from mapproxy.test.http import mock_httpd
from mapproxy.test.helper import capture


TESTSERVER_ADDRESS = ("127.0.0.1", 56413)
TESTSERVER_URL = "http://%s:%s" % TESTSERVER_ADDRESS
CAPABILITIES111_FILE = os.path.join(
    os.path.dirname(__file__), "fixture", "util_wms_capabilities111.xml"
)
CAPABILITIES130_FILE = os.path.join(
    os.path.dirname(__file__), "fixture", "util_wms_capabilities130.xml"
)
SERVICE_EXCEPTION_FILE = os.path.join(
    os.path.dirname(__file__), "fixture", "util_wms_capabilities_service_exception.xml"
)


class TestUtilWMSCapabilities(object):

    def setup_method(self):
        self.client = HTTPClient()
        self.args = ["command_dummy", "--host", TESTSERVER_URL + "/service"]

    def test_http_error(self):
        self.args = ["command_dummy", "--host", "http://foo.doesnotexist"]
        with capture() as (out, err):
            with pytest.raises(SystemExit):
                wms_capabilities_command(self.args)
        assert err.getvalue().startswith("ERROR:")

        self.args[2] = "/no/valid/url"
        with capture() as (out, err):
            with pytest.raises(SystemExit):
                wms_capabilities_command(self.args)
        assert err.getvalue().startswith("ERROR:")

    def test_request_not_parsable(self):
        with mock_httpd(
            TESTSERVER_ADDRESS,
            [
                (
                    {
                        "path": "/service?request=GetCapabilities&version=1.1.1&service=WMS",
                        "method": "GET",
                    },
                    {"status": "200", "body": ""},
                )
            ],
        ):
            with capture() as (out, err):
                with pytest.raises(SystemExit):
                    wms_capabilities_command(self.args)
            error_msg = err.getvalue().rsplit("-" * 80, 1)[1].strip()
            assert error_msg.startswith("Could not parse the document")

    def test_service_exception(self):
        self.args = [
            "command_dummy",
            "--host",
            TESTSERVER_URL + "/service?request=GetCapabilities",
        ]
        with open(SERVICE_EXCEPTION_FILE, "rb") as fp:
            capabilities_doc = fp.read()
            with mock_httpd(
                TESTSERVER_ADDRESS,
                [
                    (
                        {
                            "path": "/service?request=GetCapabilities&version=1.1.1&service=WMS",
                            "method": "GET",
                        },
                        {"status": "200", "body": capabilities_doc},
                    )
                ],
            ):
                with capture() as (out, err):
                    with pytest.raises(SystemExit):
                        wms_capabilities_command(self.args)
                error_msg = err.getvalue().rsplit("-" * 80, 1)[1].strip()
                assert "Not a capabilities document" in error_msg

    def test_parse_capabilities(self):
        self.args = [
            "command_dummy",
            "--host",
            TESTSERVER_URL + "/service?request=GetCapabilities",
            "--version",
            "1.1.1",
        ]
        with open(CAPABILITIES111_FILE, "rb") as fp:
            capabilities_doc = fp.read()
            with mock_httpd(
                TESTSERVER_ADDRESS,
                [
                    (
                        {
                            "path": "/service?request=GetCapabilities&version=1.1.1&service=WMS",
                            "method": "GET",
                        },
                        {"status": "200", "body": capabilities_doc},
                    )
                ],
            ):
                with capture() as (out, err):
                    wms_capabilities_command(self.args)
                lines = out.getvalue().split("\n")
                assert lines[1].startswith("Capabilities Document Version 1.1.1")

    def test_parse_130capabilities(self):
        self.args = [
            "command_dummy",
            "--host",
            TESTSERVER_URL + "/service?request=GetCapabilities",
            "--version",
            "1.3.0",
        ]
        with open(CAPABILITIES130_FILE, "rb") as fp:
            capabilities_doc = fp.read()
            with mock_httpd(
                TESTSERVER_ADDRESS,
                [
                    (
                        {
                            "path": "/service?request=GetCapabilities&version=1.3.0&service=WMS",
                            "method": "GET",
                        },
                        {"status": "200", "body": capabilities_doc},
                    )
                ],
            ):
                with capture() as (out, err):
                    wms_capabilities_command(self.args)
                lines = out.getvalue().split("\n")
                assert lines[1].startswith("Capabilities Document Version 1.3.0")

    def test_key_error(self):
        self.args = [
            "command_dummy",
            "--host",
            TESTSERVER_URL + "/service?request=GetCapabilities",
        ]
        with open(CAPABILITIES111_FILE, "rb") as fp:
            capabilities_doc = fp.read()
            capabilities_doc = capabilities_doc.replace(b"minx", b"foo")
            with mock_httpd(
                TESTSERVER_ADDRESS,
                [
                    (
                        {
                            "path": "/service?request=GetCapabilities&version=1.1.1&service=WMS",
                            "method": "GET",
                        },
                        {"status": "200", "body": capabilities_doc},
                    )
                ],
            ):
                with capture() as (out, err):
                    with pytest.raises(SystemExit):
                        wms_capabilities_command(self.args)

                assert err.getvalue().startswith("XML-Element has no such attribute")



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from codecs import decode
import copy
import json

from functools import reduce
from io import StringIO, BytesIO
from typing import List, Union

from lxml import etree, html


class FeatureInfoDoc(object):
    content_type = None
    content: Union[str, bytes] = None

    def as_etree(self):
        raise NotImplementedError()

    def as_string(self):
        raise NotImplementedError()


class TextFeatureInfoDoc(FeatureInfoDoc):
    info_type = "text"

    def __init__(self, content: Union[str, bytes]):
        self.content = content

    def as_string(self) -> str:
        return self.content if isinstance(self.content, str) else decode(self.content)

    @classmethod
    def combine(cls, docs: List[FeatureInfoDoc]):
        result_content = [doc.as_string() for doc in docs]
        return cls("\n".join(result_content))


class XMLFeatureInfoDoc(FeatureInfoDoc):
    info_type = "xml"
    defaultEncoding = "UTF-8"
    _etree = None

    def __init__(self, content: Union[str, bytes]):
        if isinstance(content, (str, bytes)):
            self.content = content
        else:
            if hasattr(content, "getroottree"):
                content = content.getroottree()
            self._etree = content
            assert hasattr(content, "getroot"), "expected etree like object"

    def as_string(self):
        if self.content is None:
            self.content = self._serialize_etree()
        return self.content if isinstance(self.content, str) else decode(self.content)

    def as_etree(self):
        if self._etree is None:
            self._etree = self._parse_content()
        return self._etree

    def _serialize_etree(self) -> str:
        encoding = self._etree.docinfo.encoding if \
            self._etree.docinfo.encoding else self.defaultEncoding
        return decode(etree.tostring(self._etree, encoding=encoding, xml_declaration=False), encoding)

    def _parse_content(self):
        if isinstance(self.content, str) and self.content.lstrip().startswith('<?xml'):
            # Convert back to bytes if it has XML declaration
            return etree.parse(BytesIO(self.content.encode('utf-8')))
        elif isinstance(self.content, str):
            return etree.parse(StringIO(self.content))
        else:
            return etree.parse(BytesIO(self.content))

    @classmethod
    def combine(cls, docs):
        if etree is None:
            return TextFeatureInfoDoc.combine(docs)
        doc = docs.pop(0)
        result_tree = copy.deepcopy(doc.as_etree())
        for doc in docs:
            tree = doc.as_etree()
            result_tree.getroot().extend(tree.getroot().iterchildren())

        return cls(result_tree)


class HTMLFeatureInfoDoc(XMLFeatureInfoDoc):
    info_type = "html"

    def _parse_content(self):
        root = html.document_fromstring(self.content)
        return root

    def _serialize_etree(self):
        encoding = self._etree.docinfo.encoding if \
            self._etree.docinfo.encoding else self.defaultEncoding
        return decode(html.tostring(self._etree, encoding=encoding), encoding)

    @classmethod
    def combine(cls, docs):
        if etree is None:
            return TextFeatureInfoDoc.combine(docs)

        doc = docs.pop(0)
        result_tree = copy.deepcopy(doc.as_etree())

        for doc in docs:
            tree = doc.as_etree()

            try:
                body = tree.body.getchildren()
            except IndexError:
                body = tree.getchildren()
            result_tree.body.extend(body)

        return cls(result_tree)


class JSONFeatureInfoDoc(FeatureInfoDoc):
    info_type = "json"

    def __init__(self, content):
        self.content = content

    def as_string(self):
        return self.content if isinstance(self.content, str) else decode(self.content)

    @classmethod
    def combine(cls, docs):
        contents = []
        for d in docs:
            result = json.loads(d.content)
            if result:
                contents.append(result)
        if not contents:
            return cls(json.dumps({}))
        combined = reduce(lambda a, b: merge_dict(a, b), contents)
        return cls(json.dumps(combined))


def merge_dict(base, other):
    """
    Return `base` dict with values from `conf` merged in.
    """
    for k, v in other.items():
        if k not in base:
            base[k] = v
        else:
            if isinstance(base[k], dict):
                merge_dict(base[k], v)
            elif isinstance(base[k], list):
                base[k].extend(v)
            else:
                base[k] = v
    return base


def create_featureinfo_doc(content, info_format):
    info_type = featureinfo_type(info_format)
    if info_type == "xml":
        return XMLFeatureInfoDoc(content)
    if info_type == "html":
        return HTMLFeatureInfoDoc(content)
    if info_type == "json":
        return JSONFeatureInfoDoc(content)

    return TextFeatureInfoDoc(content)


def featureinfo_type(info_format):
    info_format = info_format.split(";", 1)[
        0
    ].strip()  # remove mime options like charset
    if info_format in ("text/xml", "application/xml",
                       "application/gml+xml", "application/vnd.ogc.gml"):
        return "xml"
    if info_format == "text/html":
        return "html"
    if info_format == "application/json":
        return "json"

    return "text"


class XSLTransformer(object):

    def __init__(self, xsltscript, info_format=None):
        self.xsltscript = xsltscript
        self.info_type = featureinfo_type(info_format or "text/xml")

    def transform(self, input_doc):
        input_tree = input_doc.as_etree()
        xslt_tree = etree.parse(self.xsltscript)
        transform = etree.XSLT(xslt_tree)
        output_tree = transform(input_tree)
        if self.info_type == "html":
            return HTMLFeatureInfoDoc(output_tree)
        else:
            return XMLFeatureInfoDoc(output_tree)

    __call__ = transform


def combine_docs(docs, transformer=None):
    """
    Combine multiple FeatureInfoDocs.

    Combines as text, if the type of the docs differ.
    Otherwise, the type specific combine is called.

    Returns the combined document and the info_type (text, xml or json).
    info_type is None if the output is transformed, as the type is dependent on the
    transformer.
    """
    if len(set(d.info_type for d in docs)) > 1:
        # more than one info_type, combine as plain text
        doc = TextFeatureInfoDoc.combine(docs)
        infotype = "text"
    else:
        # all same type, combine with type specific handler
        infotype = docs[0].info_type
        doc = docs[0].combine(docs)

        if transformer:
            doc = transformer(doc)
            infotype = None  # defined by transformer
    return doc.as_string(), infotype



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import resources as importlib_resources
import copy
import json
import mimetypes
import os

import mapproxy.service as service_package
from mapproxy.config.config import base_config
from mapproxy.request.base import Request
from mapproxy.response import Response
from mapproxy.service.base import Server
from mapproxy.service.ogcapi.constants import (
    FORMAT_TYPES,
    F_JSON,
    F_HTML,
    F_PNG,
    F_JPEG,
)
from mapproxy.util.jinja2_templates import render_j2_template

import logging

log = logging.getLogger(__name__)


def static_filename(name):
    if base_config().template_dir:
        return os.path.join(base_config().template_dir, name)
    else:
        return (
            importlib_resources.files(service_package.__package__)
            .joinpath("templates")
            .joinpath(name)
        )


class OGCAPIException(Exception):
    def __init__(self, type, detail, status):
        Exception.__init__(self)
        self.type = type
        self.detail = detail
        self.status = status


class OGCAPIServer(Server):
    names = ("ogcapi",)

    def __init__(
        self,
        root_layer,
        enable_tiles=True,
        enable_maps=True,
        attribution=None,
        metadata={},
        image_formats=None,
        max_tile_age=None,
        on_error="raise",
        concurrent_layer_renderer=1,
        max_output_pixels=None,
        grid_configs=None,
        map_srs=None,
        default_dataset_layers=None,
    ):
        Server.__init__(self)
        self.enable_tiles = enable_tiles
        self.enable_maps = enable_maps
        self.image_formats = image_formats
        self.layers = root_layer.child_layers()
        self.attribution = attribution
        self.metadata = metadata if metadata else {}
        if "identification" not in self.metadata:
            self.metadata["identification"] = {}
        self.max_tile_age = max_tile_age
        self.on_error = on_error
        self.concurrent_layer_renderer = concurrent_layer_renderer
        self.max_output_pixels = max_output_pixels
        self.max_width = None
        self.max_height = None
        if max_output_pixels:
            self.max_width = int(max_output_pixels**0.5) * 10
            self.max_height = self.max_width
        self.grid_configs = {}
        if grid_configs:
            for name in grid_configs:
                conf = grid_configs[name]
                if name == "GLOBAL_WEBMERCATOR":
                    name = "WebMercatorQuad"
                self.grid_configs[name] = conf
        self.map_srs = map_srs if map_srs else []
        self.default_dataset_layers = default_dataset_layers
        self.response_headers = {}
        self.log = log

    def handle(self, req: Request):
        assert req.path.startswith("/ogcapi")

        if req.path.startswith("/ogcapi/static/") or req.path.startswith(
            "/ogcapi/demo/static/"
        ):
            if ".." in req.path:
                return Response("file not found", content_type="text/plain", status=404)
            filename = req.path.lstrip("/")
            filename = static_filename(filename)
            if not os.path.isfile(filename):
                return Response("file not found", content_type="text/plain", status=404)
            type, encoding = mimetypes.guess_type(filename)
            return Response(open(filename, "rb"), content_type=type)

        path_components = req.path.strip("/").split("/")[1:]
        log.debug(f"Handle request: {path_components} {req.args}")

        map_filenames = ["map", "map.png", "map.jpg", "map.jpeg"]

        try:
            if len(path_components) == 0:
                from mapproxy.service.ogcapi.landing_page import landing_page

                return landing_page(self, req)

            if len(path_components) == 1 and path_components[0] == "conformance":
                from mapproxy.service.ogcapi.conformance import conformance

                return conformance(self, req)

            if len(path_components) == 1 and path_components[0] == "api":
                from mapproxy.service.ogcapi.api import api

                return api(self, req)

            if len(path_components) == 1 and path_components[0] == "collections":
                from mapproxy.service.ogcapi.collections import collections

                return collections(self, req)

            if len(path_components) == 2 and path_components[0] == "collections":
                from mapproxy.service.ogcapi.collections import collection

                return collection(self, req, path_components[1])

            if (
                self.enable_maps
                and len(path_components) == 3
                and path_components[0] == "collections"
                and path_components[2] in map_filenames
            ):
                from mapproxy.service.ogcapi.map import get_map

                return get_map(self, req, path_components[1], path_components[2])

            if (
                self.enable_maps
                and len(path_components) == 1
                and path_components[0] in map_filenames
            ):
                from mapproxy.service.ogcapi.map import get_map

                return get_map(self, req, None, path_components[0])

            if (
                self.enable_tiles
                and len(path_components) == 1
                and path_components[0] == "tileMatrixSets"
            ):
                from mapproxy.service.ogcapi.tilematrixsets import tilematrixsets

                return tilematrixsets(self, req)

            if (
                self.enable_tiles
                and len(path_components) == 2
                and path_components[0] == "tileMatrixSets"
            ):
                from mapproxy.service.ogcapi.tilematrixsets import tilematrixset

                return tilematrixset(self, req, path_components[1])

            if (
                self.enable_tiles
                and len(path_components) == 4
                and path_components[0] == "collections"
                and path_components[2] == "map"
                and path_components[3] == "tiles"
            ):
                from mapproxy.service.ogcapi.tilesets import tilesets

                return tilesets(self, req, path_components[1])

            if (
                self.enable_tiles
                and len(path_components) == 5
                and path_components[0] == "collections"
                and path_components[2] == "map"
                and path_components[3] == "tiles"
            ):
                from mapproxy.service.ogcapi.tilesets import tileset

                return tileset(self, req, path_components[1], path_components[4])

            if (
                self.enable_tiles
                and len(path_components) == 8
                and path_components[0] == "collections"
                and path_components[2] == "map"
                and path_components[3] == "tiles"
            ):
                from mapproxy.service.ogcapi.tile import tile

                return tile(
                    self,
                    req,
                    path_components[1],
                    path_components[4],
                    path_components[5],
                    path_components[6],
                    path_components[7],
                )

            if (
                self.enable_tiles
                and len(path_components) == 2
                and path_components[0] == "map"
                and path_components[1] == "tiles"
            ):
                from mapproxy.service.ogcapi.tilesets import tilesets

                return tilesets(self, req, None)

            if (
                self.enable_tiles
                and len(path_components) == 3
                and path_components[0] == "map"
                and path_components[1] == "tiles"
            ):
                from mapproxy.service.ogcapi.tilesets import tileset

                return tileset(self, req, None, path_components[2])

            if (
                self.enable_tiles
                and len(path_components) == 6
                and path_components[0] == "map"
                and path_components[1] == "tiles"
            ):
                from mapproxy.service.ogcapi.tile import tile

                return tile(
                    self,
                    req,
                    None,
                    path_components[2],
                    path_components[3],
                    path_components[4],
                    path_components[5],
                )

        except OGCAPIException as e:
            json_resp = {
                "title": "OGCAPI",
                "type": e.type,
                "status": e.status,
                "detail": e.detail,
            }
            return self.create_json_or_html_response(
                req, json_resp, "exception.html", status=e.status
            )

    def create_href(self, req: Request, resource):
        return req.host_url[:-1] + resource

    def is_html_req(self, req):
        return (req.args.get("f", None) == F_HTML) or (
            FORMAT_TYPES[F_HTML] in req.accept_header
            and req.args.get("f", None) != F_JSON
        )

    def get_pygeoapi_config(self, req):
        config = {}
        config["server"] = {}
        config["server"]["url"] = req.host_url + "ogcapi"
        config["server"]["encoding"] = "utf-8"
        config["metadata"] = self.metadata
        return config

    def create_json_or_html_response(
        self, req: Request, json_resp: dict, html_page: str, status=200, headers={}
    ):
        headers = copy.copy(headers)
        headers.update(self.response_headers)

        if self.is_html_req(req):
            content = render_j2_template(
                self.get_pygeoapi_config(req),
                service_package.__package__,
                "ogcapi",
                html_page,
                json_resp,
            )
            return Response(
                content,
                content_type=FORMAT_TYPES[F_HTML],
                headers=headers,
                status=status,
            )
        else:
            return Response(
                json.dumps(json_resp, indent=2),
                content_type=FORMAT_TYPES[F_JSON],
                headers=headers,
                status=status,
            )

    @staticmethod
    def exception(type, detail, status=400):
        return OGCAPIException(type, detail, status)

    @staticmethod
    def collection_not_found():
        return OGCAPIException("Not Found", "Collection not found", status=404)

    @staticmethod
    def unknown_query_parameter(param_name: str):
        return OGCAPIException(
            "Bad Request", f"Unknown query parameter {param_name}", status=400
        )

    @staticmethod
    def invalid_parameter(msg):
        return OGCAPIException("Invalid Parameter", msg, status=400)


def get_image_format(req: Request, map_filename: str):
    """Return the image format (F_PNG/F_JPEG) from the request query parameter
    "f", the Accept header or the extension of the request path.
    """
    f = req.args.get("f", None)
    if map_filename.endswith(".png"):
        format = F_PNG
    elif map_filename.endswith(".jpg") or map_filename.endswith(".jpeg"):
        format = F_JPEG
    elif "." in map_filename:
        pos = map_filename.find(".") + 1
        format = map_filename[pos:]
        raise OGCAPIServer.exception(
            "Not Acceptable", f"Unsupported image format {format}", status=406
        )
    elif f == F_PNG:
        format = F_PNG
    elif f == F_JPEG:
        format = F_JPEG
    elif f:
        raise OGCAPIServer.exception(
            "Not Acceptable", f"Unsupported image format {f}", status=406
        )
    elif FORMAT_TYPES[F_PNG] in req.accept_header:
        format = F_PNG
    elif FORMAT_TYPES[F_JPEG] in req.accept_header:
        format = F_JPEG
    elif req.accept_header and "*/*" not in req.accept_header:
        raise OGCAPIServer.exception(
            "Not Acceptable",
            f"Unsupported image format {req.accept_header}",
            status=406,
        )
    else:
        format = F_PNG

    return format



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import shutil
import contextlib

import pytest

from mapproxy.script.export import export_command
from mapproxy.test.image import tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.helper import capture


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixture")


@contextlib.contextmanager
def tile_server(tile_coords):
    with tmp_image((256, 256), format="jpeg") as img:
        img = img.read()
    expected_reqs = []
    for tile in tile_coords:
        expected_reqs.append(
            (
                {"path": r"/tiles/%d/%d/%d.png" % (tile[2], tile[0], tile[1])},
                {"body": img, "headers": {"content-type": "image/png"}},
            )
        )
    with mock_httpd(("localhost", 42423), expected_reqs, unordered=True):
        yield


class TestUtilExport(object):

    def setup_method(self):
        self.dir = tempfile.mkdtemp()
        self.dest = os.path.join(self.dir, "dest")
        self.mapproxy_conf_name = "mapproxy_export.yaml"
        shutil.copy(os.path.join(FIXTURE_DIR, self.mapproxy_conf_name), self.dir)
        self.mapproxy_conf_file = os.path.join(self.dir, self.mapproxy_conf_name)
        self.args = ["command_dummy", "-f", self.mapproxy_conf_file]

    def teardown_method(self):
        shutil.rmtree(self.dir)

    def test_config_not_found(self):
        self.args = ["command_dummy", "-f", "foo.bar"]
        with capture() as (out, err):
            try:
                export_command(self.args)
            except SystemExit as ex:
                assert ex.code != 0
            else:
                assert False, "export command did not exit"
        assert err.getvalue().startswith("ERROR:")

    def test_no_fetch_missing_tiles(self):
        self.args += [
            "--grid",
            "GLOBAL_MERCATOR",
            "--dest",
            self.dest,
            "--levels",
            "0",
            "--source",
            "tms_cache",
        ]
        with capture() as (out, err):
            export_command(self.args)

        assert os.listdir(self.dest) == ["tile_locks"]

    def test_fetch_missing_tiles(self):
        self.args += [
            "--grid",
            "GLOBAL_MERCATOR",
            "--dest",
            self.dest,
            "--levels",
            "0,1",
            "--source",
            "tms_cache",
            "--fetch-missing-tiles",
        ]
        with tile_server([(0, 0, 0), (0, 0, 1), (0, 1, 1), (1, 0, 1), (1, 1, 1)]):
            with capture() as (out, err):
                export_command(self.args)

        assert os.path.exists(os.path.join(self.dest, "tile_locks"))
        assert os.path.exists(os.path.join(self.dest, "0", "0", "0.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "0", "0.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "0", "1.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "1", "0.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "1", "1.png"))

    def test_force(self):
        self.args += [
            "--grid",
            "GLOBAL_MERCATOR",
            "--dest",
            self.dest,
            "--levels",
            "0",
            "--source",
            "tms_cache",
        ]
        with capture() as (out, err):
            export_command(self.args)

        with capture() as (out, err):
            with pytest.raises(SystemExit):
                export_command(self.args)

        with capture() as (out, err):
            export_command(self.args + ["--force"])

    def test_invalid_grid_definition(self):
        self.args += [
            "--grid",
            "foo=1",
            "--dest",
            self.dest,
            "--levels",
            "0",
            "--source",
            "tms_cache",
        ]
        with capture() as (out, err):
            with pytest.raises(SystemExit):
                export_command(self.args)
            assert "foo" in err.getvalue()

    def test_custom_grid(self):
        self.args += [
            "--grid",
            "base=GLOBAL_MERCATOR min_res=100000",
            "--dest",
            self.dest,
            "--levels",
            "1",
            "--source",
            "tms_cache",
            "--fetch-missing-tiles",
        ]
        with tile_server(
            [
                (0, 3, 2),
                (1, 3, 2),
                (2, 3, 2),
                (3, 3, 2),
                (0, 2, 2),
                (1, 2, 2),
                (2, 2, 2),
                (3, 2, 2),
                (0, 1, 2),
                (1, 1, 2),
                (2, 1, 2),
                (3, 1, 2),
                (0, 0, 2),
                (1, 0, 2),
                (2, 0, 2),
                (3, 0, 2),
            ]
        ):
            with capture() as (out, err):
                export_command(self.args)

        assert os.path.exists(os.path.join(self.dest, "tile_locks"))
        assert os.path.exists(os.path.join(self.dest, "1", "0", "0.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "3", "3.png"))

    def test_coverage(self):
        self.args += [
            "--grid",
            "GLOBAL_MERCATOR",
            "--dest",
            self.dest,
            "--levels",
            "0..2",
            "--source",
            "tms_cache",
            "--fetch-missing-tiles",
            "--coverage",
            "10,10,20,20",
            "--srs",
            "EPSG:4326",
        ]
        with tile_server([(0, 0, 0), (1, 1, 1), (2, 2, 2)]):
            with capture() as (out, err):
                export_command(self.args)

        assert os.path.exists(os.path.join(self.dest, "tile_locks"))
        assert os.path.exists(os.path.join(self.dest, "0", "0", "0.png"))
        assert os.path.exists(os.path.join(self.dest, "1", "1", "1.png"))
        assert os.path.exists(os.path.join(self.dest, "2", "2", "2.png"))



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.template import bunch

__all__ = ['indent', 'bunch', 'wms100format', 'wms100info_format', 'wms111metadatatype']


def indent(text, n=2):
    return '\n'.join(' '*n + line for line in text.split('\n'))


def wms100format(format):
    """
    >>> wms100format('image/png')
    'PNG'
    >>> wms100format('image/GeoTIFF')
    """
    _mime_class, sub_type = format.split('/')
    sub_type = sub_type.upper()
    if sub_type in ['PNG', 'TIFF', 'GIF', 'JPEG']:
        return sub_type
    else:
        return None


def wms100info_format(format):
    """
    >>> wms100info_format('text/html')
    'MIME'
    >>> wms100info_format('application/vnd.ogc.gml')
    'GML.1'
    """
    if format in ('application/vnd.ogc.gml', 'text/xml'):
        return 'GML.1'
    return 'MIME'


def wms111metadatatype(type):
    if type == 'ISO19115:2003':
        return 'TC211'
    if type == 'FGDC:1998':
        return 'FGDC'



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from mapproxy.test.image import tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "cache_grid_names.yaml"


class TestCacheGridNames(SysTest):

    def test_tms_capabilities(self, app):
        resp = app.get("/tms/1.0.0/")
        assert "Cached Layer" in resp
        assert "wms_cache/utm32n" in resp
        assert "wms_cache_utm32n" not in resp
        xml = resp.lxml
        assert xml.xpath("count(//TileMap)") == 2

    def test_tms_layer_capabilities(self, app):
        resp = app.get("/tms/1.0.0/wms_cache/utm32n")
        assert "Cached Layer" in resp
        assert "wms_cache/utm32n" in resp
        assert "wms_cache_utm32n" not in resp
        xml = resp.lxml
        assert xml.xpath("count(//TileSet)") == 12

    def test_kml(self, app):
        resp = app.get("/kml/wms_cache/utm32n/4/2/2.kml")
        assert b"wms_cache/utm32n" in resp.body

    def test_get_tile(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A25832&styles="
                    "&VERSION=1.1.1&&WIDTH=256"
                    "&BBOX=283803.311362,5609091.90862,319018.942566,5644307.53982"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/wms_cache/utm32n/4/2/2.jpeg")
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "wms_cache/utm32n/04/000/000/002/000/000/002.jpeg"
        ).check()

    def test_get_tile_no_grid_name(self, app, cache_dir):
        # access tiles with grid name from TMS but cache still uses old SRS-code path
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A25832&styles="
                    "&VERSION=1.1.1&WIDTH=256"
                    "&BBOX=283803.311362,5609091.90862,319018.942566,5644307.53982"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/wms_cache_no_grid_name/utm32n/4/2/2.jpeg")
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "wms_cache_no_grid_name_EPSG25832/04/000/000/002/000/000/002.jpeg"
        ).check()



# -:- encoding: utf8 -:-
# Copyright (c) 2011, Oliver Tonnhofer <olt@omniscale.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import unittest

from ..validator import validate, ValidationError, SpecError
from ..spec import required, one_of, number, recursive, type_spec, anything


def raises(exception):
    def wrapper(f):
        def _wrapper(self):
            try:
                f(self)
            except exception:
                pass
            else:
                raise AssertionError('expected exception %s', exception)
    return wrapper


class TestSimpleDict(unittest.TestCase):
    def test_validate_simple_dict(self):
        spec = {'hello': 1, 'world': True}
        validate(spec, {'hello': 34, 'world': False})

    @raises(ValidationError)
    def test_invalid_key(self):
        spec = {'world': True}
        validate(spec, {'world_foo': False})

    def test_empty_data(self):
        spec = {'world': 1}
        validate(spec, {})

    @raises(ValidationError)
    def test_invalid_value(self):
        spec = {'world': 1}
        validate(spec, {'world_foo': False})

    @raises(ValidationError)
    def test_missing_required_key(self):
        spec = {required('world'): 1}
        validate(spec, {})

    def test_valid_one_of(self):
        spec = {'hello': one_of(1, bool())}
        validate(spec, {'hello': 129})
        validate(spec, {'hello': True})

    @raises(ValidationError)
    def test_invalid_one_of(self):
        spec = {'hello': one_of(1, False)}
        validate(spec, {'hello': []})

    def test_instances_and_types(self):
        spec = {'str()': str(), 'string_type': str, 'int': int, 'int()': int()}
        validate(spec, {'str()': 'str', 'string_type': u'☃', 'int': 1, 'int()': 1})


class TestLists(unittest.TestCase):
    def test_list(self):
        spec = [1]
        validate(spec, [1, 2, 3, 4, -9])

    def test_empty_list(self):
        spec = [1]
        validate(spec, [])

    @raises(ValidationError)
    def test_invalid_item(self):
        spec = [1]
        validate(spec, [1, 'hello'])


class TestNumber(unittest.TestCase):
    def check_valid(self, spec, data):
        validate(spec, data)

    def test_numbers(self):
        spec = number()
        for i in (0, 1, 23e999, int(10e20), 23.1, -0.0000000001):
            self.check_valid(spec, i)


class TestNested(unittest.TestCase):
    def check_valid(self, spec, data):
        validate(spec, data)

    def check_invalid(self, spec, data):
        try:
            validate(spec, data)
        except ValidationError:
            pass
        else:
            assert False, "expected ValidationError"

    def test_dict(self):
        spec = {
            'globals': {
                'image': {
                    'format': {
                        'png': {
                            'mode': 'RGB',
                        }
                    },
                },
                'cache': {
                    'base_dir': '/path/to/foo'
                }
            }
        }

        self.check_valid(spec, {'globals': {'image': {'format': {'png': {'mode': 'P'}}}}})
        self.check_valid(spec, {'globals': {'image': {'format': {'png': {'mode': 'P'}}},
                                            'cache': {'base_dir': '/somewhere'}}})
        self.check_invalid(spec, {'globals': {'image': {'foo': {'png': {'mode': 'P'}}}}})
        self.check_invalid(spec, {'globals': {'image': {'png': {'png': {'mode': 1}}}}})

    def test_errors_in_unicode_keys(self):
        # should not raise UnicodeEncodeError
        spec = {
            anything(): str(),
        }
        self.check_invalid(spec, {u'globalü': 12})


class TestRecursive(unittest.TestCase):
    def test(self):
        spec = recursive({'hello': str(), 'more': recursive()})
        validate(spec, {'hello': 'world', 'more': {'hello': 'foo', 'more': {'more': {}}}})

    def test_multiple(self):
        spec = {'a': recursive({'hello': str(), 'more': recursive()}), 'b': recursive({'foo': recursive()})}
        validate(spec, {'b': {'foo': {'foo': {}}}})
        validate(spec, {'a': {'hello': 'world', 'more': {'hello': 'foo', 'more': {'more': {}}}}})
        validate(spec, {'b': {'foo': {'foo': {}}},
                        'a': {'hello': 'world', 'more': {'hello': 'foo', 'more': {'more': {}}}}})

    @raises(SpecError)
    def test_without_spec(self):
        spec = {'a': recursive()}
        validate(spec, {'a': {'a': {}}})


class TestTypeSpec(unittest.TestCase):
    def test(self):
        spec = type_spec('type', {'foo': {'alpha': str()}, 'bar': {'one': 1, 'two': str()}})
        validate(spec, {'type': 'foo', 'alpha': 'yes'})
        validate(spec, {'type': 'bar', 'one': 2})

    def test_missing_type(self):
        spec = type_spec('type', {'foo': {'alpha': str()}, 'bar': {'one': 1, 'two': str()}})
        try:
            validate(spec, {'alpha': 'yes'})
        except ValidationError as ex:
            assert "'type' not in ." in ex.errors[0]
        else:
            assert False

    def test_unknown_type(self):
        spec = type_spec('type', {'foo': {'alpha': str()}, 'bar': {'one': 1, 'two': str()}})
        try:
            validate(spec, {'type': 'baz', 'alpha': 'yes'})
        except ValidationError as ex:
            assert "unknown type value 'baz' in ." in ex.errors[0], ex
        else:
            assert False

    def test_no_type_dict(self):
        spec = {'dict': type_spec('type', {'foo': {'alpha': str()}, 'bar': {'one': 1, 'two': str()}})}
        try:
            validate(spec, {'dict': None})
        except ValidationError as ex:
            assert "dict is empty" in ex.errors[0], ex
        else:
            assert False


class TestErrors(unittest.TestCase):
    def test_invalid_types(self):
        spec = {'str': str, 'str()': str(), 'string_type': str, '1': 1, 'int': int}
        try:
            validate(spec, {'str': 1, 'str()': 1, 'string_type': 1, '1': 'a', 'int': 'int'})
        except ValidationError as ex:
            ex.errors.sort()
            assert ex.errors[0] == "'a' in 1 not of type int"
            assert ex.errors[1] == "'int' in int not of type int"
            assert ex.errors[2] == '1 in str not of type str'
            assert ex.errors[3] == '1 in str() not of type str'
            assert ex.errors[4] == '1 in string_type not of type str'
        else:
            assert False

    def test_invalid_key(self):
        spec = {'world': {'europe': {}}}
        try:
            validate(spec, {'world': {'europe': {'germany': 1}}})
        except ValidationError as ex:
            assert 'world.europe' in str(ex)
        else:
            assert False

    def test_invalid_list_item(self):
        spec = {'numbers': [number()]}
        try:
            validate(spec, {'numbers': [1, 2, 3, 'foo']})
        except ValidationError as ex:
            assert 'numbers[3] not of type number' in str(ex), str(ex)
        else:
            assert False

    def test_multiple_invalid_list_items(self):
        spec = {'numbers': [number()]}
        try:
            validate(spec, {'numbers': [1, True, 3, 'foo']})
        except ValidationError as ex:
            assert '2 validation errors' in str(ex), str(ex)
            assert 'numbers[1] not of type number' in ex.errors[0]
            assert 'numbers[3] not of type number' in ex.errors[1]
        else:
            assert False

    def test_error_in_non_string_key(self):
        spec = {1: bool()}
        try:
            validate(spec, {1: 'not a bool'})
        except ValidationError as ex:
            assert "'not a bool' in 1 not of type bool" in ex.errors[0]
        else:
            assert False

    def test_error_in_non_string_key_with_anything_key_spec(self):
        spec = {anything(): bool()}
        try:
            validate(spec, {1: 'not a bool'})
        except ValidationError as ex:
            assert "'not a bool' in 1 not of type bool" in ex.errors[0]
        else:
            assert False


def test_one_of_with_custom_types():
    # test for fixed validation of one_of specs with values that are
    # not lists or dicts (e.g. recursive)
    spec = one_of([str], recursive({required('foo'): str}))
    validate(spec, ['foo', 'bar'])
    validate(spec, {'foo': 'bar'})
    try:
        validate(spec, {'nofoo': 'bar'})
    except ValidationError as ex:
        assert "missing 'foo'" in ex.errors[0]
    else:
        assert False


if __name__ == '__main__':
    unittest.main()



# This file is part of the MapProxy project.
# Copyright (C) 2022 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
from io import BytesIO

import pytest

from mapproxy.request.wms import WMS111MapRequest
from mapproxy.test.image import is_png, create_tmp_image
from mapproxy.test.system import SysTest

try:
    from mapproxy.cache.azureblob import AzureBlobCache
except ImportError:
    AzureBlobCache = None


def azureblob_connection():
    # Use default storage account of Azurite emulator
    host = os.environ['MAPPROXY_TEST_AZURE_BLOB']
    return 'DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=' \
           'Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr' \
           '/KBHBeksoGMGw==;BlobEndpoint=' + host + '/devstoreaccount1;'


def azure_container_client(connection_string, container_name):
    return AzureBlobCache(
        base_path="",
        file_ext="",
        connection_string=connection_string,
        container_name=container_name).container_client


@pytest.fixture(scope="module")
def config_file():
    return "cache_azureblob.yaml"


@pytest.fixture(scope="module")
def azureblob_containers():
    container = azure_container_client(azureblob_connection(), 'default-container')
    if not container.exists():
        container.create_container()
    container = azure_container_client(azureblob_connection(), 'tiles')
    if not container.exists():
        container.create_container()

    yield


@pytest.mark.skipif(not AzureBlobCache or not os.environ.get('MAPPROXY_TEST_AZURE_BLOB'),
                    reason="azure-storage-blob package and MAPPROXY_TEST_AZURE_BLOB env required")
@pytest.mark.usefixtures("azureblob_containers")
class TestAzureBlobCache(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-150,-40,-140,-30",
                width="100",
                height="100",
                layers="default",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )

    def test_get_map_cached(self, app):
        tile = create_tmp_image((256, 256))
        container = azure_container_client(azureblob_connection(), 'default-container')
        container.upload_blob(
            name="default_cache/WebMerc/4/1/9.png",
            data=BytesIO(tile),
            overwrite=True
        )
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_cached_quadkey(self, app):
        tile = create_tmp_image((256, 256))
        container = azure_container_client(azureblob_connection(), 'tiles')
        container.upload_blob(
            name="quadkeytiles/2003.png",
            data=BytesIO(tile),
            overwrite=True
        )
        self.common_map_req.params.layers = "quadkey"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_cached_reverse_tms(self, app):
        tile = create_tmp_image((256, 256))
        container = azure_container_client(azureblob_connection(), 'tiles')
        container.upload_blob(
            name="reversetiles/9/1/4.png",
            data=BytesIO(tile),
            overwrite=True
        )
        self.common_map_req.params.layers = "reverse"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)



import os

from mapproxy.test.helper import TempDir, assert_files_in_dir, ChangeWorkingDir, assert_permissions
from mapproxy.util.fs import ensure_directory


class TestFs:
    def test_ensure_directory_absolute(self):
        with TempDir() as dir_name:
            file_path = os.path.join(dir_name, 'a/b/c')
            ensure_directory(file_path)
            assert_files_in_dir(dir_name, ['a'])
            assert_files_in_dir(os.path.join(dir_name, 'a'), ['b'])
            assert_files_in_dir(os.path.join(dir_name, 'a/b'), [])

    def test_ensure_directory_relative(self):
        with TempDir() as dir_name:
            with ChangeWorkingDir(dir_name):
                ensure_directory('./a/b/c')
                assert_files_in_dir(dir_name, ['a'])
                assert_files_in_dir(os.path.join(dir_name, 'a'), ['b'])
                assert_files_in_dir(os.path.join(dir_name, 'a/b'), [])

    def test_ensure_directory_permissions(self):
        with TempDir() as dir_name:
            file_path = os.path.join(dir_name, 'a/b')
            desired_permissions = '700'
            ensure_directory(file_path, directory_permissions=desired_permissions)
            assert_files_in_dir(dir_name, ['a'])
            assert_permissions(os.path.join(dir_name, 'a'), desired_permissions)
